from ...utils import PlatformCommand


class DriveAppCommand(PlatformCommand):
    """DriveApp related commands."""

    from .create import CreateCommand
    from .delete import DeleteCommand
    from .update import UpdateCommand



from ...utils import (
    PlatformCommand,
    valid_path
)
from ..utils import EngagePlatformManager


class CreateFromCommand(PlatformCommand):
    """Create an EngageAppVersion from an other one."""

    def setup(self, subparsers):
        parser = super(CreateFromCommand, self).setup(subparsers)
        parser.add_argument('--from', dest="origin", required=True, type=str, help="EngageAppVersion id used as base")
        parser.add_argument('-w', '--workflow', default=None, type=valid_path, help="Path to the workflow yaml file")
        parser.add_argument('-c', '--custom_nodes', default=None, type=valid_path, help="Path to the custom nodes python file")
        parser.add_argument('-p', '--base_major_version', default=None, type=str, help="Previous major EngageAppVersion version")
        parser.add_argument('-r', '--recognition-version-ids', nargs="*", type=int,
                            help="List of Recognition Version Id, one for each Recognition Spec in the App", default=[])
        return parser

    def run(self,
            origin,
            base_major_version,
            workflow,
            custom_nodes,
            recognition_version_ids,
            **kwargs):

        return EngagePlatformManager().create_app_version_from(
            origin,
            base_major_version,
            workflow,
            custom_nodes,
            recognition_version_ids
        )



from ...utils import Command


class WorkOrderCommand(Command):
    """
        Work order related commands
    """

    from .create import CreateCommand
    from .status import StatusCommand
    from .inference import InferCommand
    from .delete import DeleteCommand



import sys
import cv2
import logging
import numpy as np
import threading
from text_unidecode import unidecode
from tqdm import tqdm

from deepomatic.cli.common import Queue, TqdmToLogger
from deepomatic.cli.exceptions import (DeepoCLICredentialsError,
                                       SendInferenceError,
                                       ResultInferenceError,
                                       ResultInferenceTimeout)
from deepomatic.cli.frame import CurrentFrames
from deepomatic.cli.input_data import InputThread, VideoInputData, get_input
from deepomatic.cli.output_data import OutputThread
from deepomatic.cli.thread_base import QUEUE_MAX_SIZE, MainLoop, Pool, Thread, Greenlet
from deepomatic.cli.workflow import get_workflow


LOGGER = logging.getLogger(__name__)


# Draw parameters
SCORE_DECIMAL_PRECISION = 4             # Prediction score decimal number precision
FONT_SCALE = 0.5                        # Size of the font we draw in the image_output
BOX_COLOR = (255, 0, 0)                 # Bounding box color (BGR)
TEXT_COLOR = (255, 255, 255)            # Text color (BGR)
TAG_TEXT_CORNER = (10, 10)              # Beginning of text tag column (pixel)
TAG_TEXT_INTERSPACE = 5                 # Vertical space between tags in tag column (pixel)


def substract_tuple(tuple1, tuple2):
    return tuple(x - y for x, y in zip(tuple1, tuple2))


def get_coordinates_from_roi(roi, width, height):
    bbox = roi['bbox']
    xmin = int(bbox['xmin'] * width)
    ymin = int(bbox['ymin'] * height)
    xmax = int(bbox['xmax'] * width)
    ymax = int(bbox['ymax'] * height)
    return (xmin, ymin, xmax, ymax)


class DrawImagePostprocessing(object):

    def __init__(self, **kwargs):
        self._draw_labels = kwargs['draw_labels']
        self._draw_scores = kwargs['draw_scores']
        self._font_scale = kwargs['font_scale']
        self._font_thickness = kwargs['font_thickness']
        self._threshold = kwargs['threshold'] if kwargs['threshold'] is not None else 0
        self._font_bg_color = kwargs['font_bg_color']

    def __call__(self, frame):
        frame.output_image = frame.image.copy()
        output_image = frame.output_image
        height = output_image.shape[0]
        width = output_image.shape[1]
        tag_drawn = 0  # Used to store the number of tags already drawn
        for pred in frame.predictions['outputs'][0]['labels']['predicted']:
            score = pred['score']
            # Build legend
            label = u''
            if self._draw_labels:
                label = pred['label_name']
            if self._draw_labels and self._draw_scores:
                label += ' '
            if self._draw_scores:
                label += str(round(score, SCORE_DECIMAL_PRECISION))

            # Make sure labels are ascii because cv2.FONT_HERSHEY_SIMPLEX doesn't support non-ascii
            label = unidecode(label)

            # Get text draw parameters
            ret, baseline = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, self._font_scale, self._font_thickness)

            if self._font_bg_color:
                background_color = tuple(self._font_bg_color)
            # Get background color depending on score with a gradient from green to red depending on the threshold set
            else:
                hsv_color = (60 * (score - self._threshold) / (1 - self._threshold), 255, 200)
                background_color = cv2.cvtColor(np.uint8([[hsv_color]]), cv2.COLOR_HSV2BGR).flatten()
                background_color = background_color.astype('float64')
            # If we have a bounding box
            roi = pred.get('roi')
            if roi is not None:
                # Retrieve coordinates
                xmin, ymin, xmax, ymax = get_coordinates_from_roi(roi, width, height)

                # Draw bounding box
                cv2.rectangle(output_image, (xmin, ymin), (xmax, ymax), BOX_COLOR, 1)

                if label != '':
                    # First get ideal corners
                    background_corner1 = (xmin, ymax + 2)
                    background_corner2 = (background_corner1[0] + ret[0], background_corner1[1] + ret[1] + baseline)
                    text_corner = (background_corner1[0], background_corner1[1] + ret[1])

                    # Then make sure they fit in the image
                    # For x-axis, simply shift the box to the left
                    # For y-axis, put the label at the top if it doesn't fit under the box
                    x_offset = max(0, background_corner2[0] - width + 1)
                    if background_corner2[1] > height - 1:
                        y_offset = background_corner2[1] - ymin + 2
                    else:
                        y_offset = 0
                    background_corner1 = substract_tuple(background_corner1, (x_offset, y_offset))
                    background_corner2 = substract_tuple(background_corner2, (x_offset, y_offset))
                    text_corner = substract_tuple(text_corner, (x_offset, y_offset))

                    # Finally draw everything
                    cv2.rectangle(output_image, background_corner1, background_corner2, background_color, -1)
                    cv2.putText(output_image, label, text_corner, cv2.FONT_HERSHEY_SIMPLEX,
                                self._font_scale, TEXT_COLOR, self._font_thickness)
            elif label != '':
                # First get ideal corners
                if tag_drawn == 0:
                    background_corner1 = TAG_TEXT_CORNER
                else:
                    background_corner1 = (TAG_TEXT_CORNER[0], background_corner2[1] + TAG_TEXT_INTERSPACE)
                background_corner2 = (background_corner1[0] + ret[0], background_corner1[1] + ret[1] + baseline)
                text_corner = (background_corner1[0], background_corner1[1] + ret[1])

                # Finally draw everything
                cv2.rectangle(output_image, background_corner1, background_corner2, background_color, -1)
                cv2.putText(output_image, label, text_corner, cv2.FONT_HERSHEY_SIMPLEX, self._font_scale, TEXT_COLOR, self._font_thickness)
                tag_drawn += 1


class BlurImagePostprocessing(object):
    def __init__(self, **kwargs):
        self._method = kwargs.get('blur_method', 'pixel')
        self._strength = int(kwargs.get('blur_strength', 10))

    def __call__(self, frame):
        frame.output_image = frame.image.copy()
        output_image = frame.output_image
        height = output_image.shape[0]
        width = output_image.shape[1]
        skipped_pred = 0
        for pred in frame.predictions['outputs'][0]['labels']['predicted']:
            # Check that we have a bounding box
            roi = pred.get('roi')
            if roi is not None:
                # Retrieve coordinates
                xmin, ymin, xmax, ymax = get_coordinates_from_roi(roi, width, height)
                if (xmax > xmin) and (ymax > ymin):
                    # Draw
                    if self._method == 'black':
                        cv2.rectangle(output_image, (xmin, ymin), (xmax, ymax), (0, 0, 0), -1)
                    elif self._method == 'gaussian':
                        rectangle = output_image[ymin:ymax, xmin:xmax]
                        rectangle = cv2.GaussianBlur(rectangle, (0, 0), self._strength)
                        output_image[ymin:ymax, xmin:xmax] = rectangle
                    elif self._method == 'pixel':
                        rectangle = output_image[ymin:ymax, xmin:xmax]
                        small = cv2.resize(rectangle, (0, 0),
                                           fx=1. / min((xmax - xmin), self._strength),
                                           fy=1. / min((ymax - ymin), self._strength))
                        rectangle = cv2.resize(small, ((xmax - xmin), (ymax - ymin)),
                                               interpolation=cv2.INTER_NEAREST)
                        output_image[ymin:ymax, xmin:xmax] = rectangle
                else:
                    skipped_pred += 1
        if skipped_pred > 0:
            LOGGER.warning("Skipped {} predictions (invalid bbox)".format(skipped_pred))


class PrepareInferenceThread(Thread):
    def process_msg(self, frame):
        try:
            _, buf = cv2.imencode('.jpg', frame.image)
        except Exception as e:
            LOGGER.error('Could not decode image for frame {}: {}'.format(frame, e))
            return None
        buf_bytes = buf.tobytes()
        frame.buf_bytes = buf_bytes
        self.current_messages.add_frame(frame)
        return frame


class SendInferenceGreenlet(Greenlet):
    def __init__(self, exit_event, input_queue, output_queue, current_messages, workflow):
        super(SendInferenceGreenlet, self).__init__(exit_event, input_queue, output_queue, current_messages)
        self.workflow = workflow
        self.push_client = workflow.new_client()

    def close(self):
        self.workflow.close_client(self.push_client)

    def process_msg(self, frame):
        try:
            frame.inference_async_result = self.workflow.infer(frame.buf_bytes, self.push_client, frame.name)
            return frame
        except SendInferenceError as e:
            self.current_messages.forget_frame(frame)
            LOGGER.error('Error sending frame {}: {}'.format(frame, e))
            return None


class ResultInferenceGreenlet(Greenlet):
    def __init__(self, exit_event, input_queue, output_queue, current_messages, workflow, **kwargs):
        super(ResultInferenceGreenlet, self).__init__(exit_event, input_queue, output_queue, current_messages)
        self.workflow = workflow
        self.threshold = kwargs.get('threshold')

    def fill_predictions(self, predictions, new_predicted, new_discarded):
        for prediction in predictions:
            if prediction['score'] >= self.threshold:
                new_predicted.append(prediction)
            else:
                new_discarded.append(prediction)

    def process_msg(self, frame):
        try:
            predictions = frame.inference_async_result.get_predictions(timeout=60)
            if self.threshold is not None:
                # Keep only predictions higher than threshold
                for output in predictions['outputs']:
                    new_predicted = []
                    new_discarded = []
                    labels = output['labels']
                    self.fill_predictions(labels['predicted'], new_predicted, new_discarded)
                    self.fill_predictions(labels['discarded'], new_predicted, new_discarded)
                    labels['predicted'] = new_predicted
                    labels['discarded'] = new_discarded

            frame.predictions = predictions
            return frame
        except ResultInferenceTimeout as e:
            self.current_messages.forget_frame(frame)
            LOGGER.error("Couldn't get predictions in {} seconds. Ignoring frame {}.".format(e.timeout, frame))
        except ResultInferenceError as e:
            self.current_messages.forget_frame(frame)
            LOGGER.error('Error getting predictions for frame {}: {}'.format(frame, e))
        return None


class InferManager(object):

    def input_loop(self, kwargs, postprocessing=None):
        # Adds smartness to fps handling
        #   1) If both input_fps and output_fps are set, then use them as is.
        #   2) If only one of the two is used, make both equal
        #   3) If none is set:
        #       * If the input is not a video, do nothing and use the default DEFAULT_FPS output value
        #       * If the input is a video, use the input fps as the output fps
        if kwargs['input_fps'] and not kwargs['output_fps']:
            kwargs['output_fps'] = kwargs['input_fps']
            LOGGER.info('Input fps of {} specified, but no output fps specified. Using same value for both.'.format(kwargs['input_fps']))
        elif kwargs['output_fps'] and not kwargs['input_fps']:
            kwargs['input_fps'] = kwargs['output_fps']
            LOGGER.info('Output fps of {} specified, but no input fps specified. Using same value for both.'.format(kwargs['output_fps']))

        # Compute inputs now to access actual input fps if it's a video
        inputs = iter(get_input(kwargs.get('input', 0), kwargs))

        # Deal with last case for fps
        if not(kwargs['input_fps']) and not(kwargs['output_fps']) and isinstance(inputs, VideoInputData):
            kwargs['input_fps'] = inputs.get_fps()
            kwargs['output_fps'] = kwargs['input_fps']
            LOGGER.info('Input fps of {} automatically detected, but no output fps specified.'
                        ' Using same value for both.'.format(kwargs['input_fps']))

        # Initialize progress bar
        frame_count = inputs.get_frame_count()
        max_value = int(frame_count) if frame_count >= 0 else None
        tqdmout = TqdmToLogger(LOGGER, level=LOGGER.getEffectiveLevel())
        pbar = tqdm(total=max_value, file=tqdmout, desc='Input processing', smoothing=0)

        # Initialize workflow for mutual use between send inference pool and result inference pool
        try:
            workflow = get_workflow(kwargs)
        except DeepoCLICredentialsError as e:
            LOGGER.error(str(e))
            sys.exit(1)

        # IMPORTANT: maxsize is important, it allows to regulate the pipeline and
        # avoid to pushes too many requests to rabbitmq when we are already waiting for many results

        nb_queue = 2  # input => prepare inference => output
        if workflow:
            nb_queue += 2  # prepare inference => send inference => result inference

        queues = [Queue(maxsize=QUEUE_MAX_SIZE) for _ in range(nb_queue)]

        exit_event = threading.Event()

        current_frames = CurrentFrames()

        pools = [
            Pool(1, InputThread, thread_args=(exit_event, None, queues[0], inputs)),
            # Encode image into jpeg
            Pool(1, PrepareInferenceThread, thread_args=(exit_event, queues[0], queues[1], current_frames)),
        ]

        if workflow:
            pools.extend([
                # Send inference
                Pool(5, SendInferenceGreenlet, thread_args=(exit_event, queues[1], queues[2], current_frames, workflow)),
                # Gather inference predictions from the worker(s)
                Pool(1, ResultInferenceGreenlet, thread_args=(exit_event, queues[2], queues[3], current_frames, workflow),
                     thread_kwargs=kwargs),
            ])

        # Output predictions
        pools.append(Pool(1, OutputThread, thread_args=(exit_event, queues[-1], None, current_frames, pbar.update, postprocessing),
                          thread_kwargs=kwargs))

        loop = MainLoop(pools, queues, pbar, exit_event, current_frames, lambda: workflow.close() if workflow else None)

        try:
            stop_asked = loop.run_forever()
        except Exception:
            loop.cleanup()
            raise

        # If the process encountered an error, the exit code is 1.
        # If the process is interrupted using SIGINT (ctrl + C) or SIGTERM, the queues are emptied and processed by the
        # threads, and the exit code is 0.
        # If SIGINT or SIGTERM is sent again during this shutdown phase, the threads are killed, and the exit code is 2.
        if exit_event.is_set():
            sys.exit(1)
        elif stop_asked >= 2:
            sys.exit(2)



from jsonschema import validate, ValidationError


class JSONSchemaType:
    STUDIO_HEADER = 'Studio-header'
    STUDIO_INPUT = 'Studio-input'
    VULCAN = 'Vulcan'
    # Define the vulcan json schema
    VULCAN_ANNOTATION_SCHEMA = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "roi": {},
                "score": {},
                "label_id": {},
                "threshold": {},
                "label_name": {}
            }
        }
    }
    VULCAN_JSON_SCHEMA = {
        "type": "array",
        "items": {
            "type": "object",
            "required": ["outputs"],
            "properties": {
                "location": {"type": "string"},
                "outputs": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "required": ["labels"],
                        "properties": {
                            "labels": {
                                "type": "object",
                                "properties": {
                                    "discarded": VULCAN_ANNOTATION_SCHEMA,
                                    "predicted": VULCAN_ANNOTATION_SCHEMA
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    # Define the studio json format
    CONCEPT_SCHEMA = {
        "type": "object",
        "required": ["name"],
        "additionalProperties": False,
        "properties": {
            "name": {
                "type": "string"
            },
            "bool": {
                "type": "boolean"
            }
        }
    }

    HEADER_SCHEMA = {
        "type": "object",
        "required": [
            "name",
            "splits",
            "views"
        ],
        "additionalProperties": False,
        "properties": {
            "name": {
                "type": "string"
            },
            "splits": {
                "type": "array",
                "items": [
                    {
                        "type": "string",
                        "enum": ["train", "val"]
                    },
                    {
                        "type": "string",
                        "enum": ["train", "val"]
                    }
                ]
            },
            "views": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": [
                        "concepts",
                        "type",
                        "name",
                        "conditions"
                    ],
                    "additionalProperties": False,
                    "properties": {
                        "name": {
                            "type": "string"
                        },
                        "type": {
                            "type": "string"
                        },
                        "concepts": {
                            "type": "array",
                            "items": CONCEPT_SCHEMA
                        },
                        "conditions": {
                            "type": "array"
                        },
                        "id": {
                            "type": ["number", "null"]
                        }
                    }
                }
            }
        }
    }

    ANNOTATION_SCHEMA = {
        "type": "array",
        "items": {
            "type": "object",
            "required": ["concepts", "region"],
            "additionalProperties": False,
            "properties": {
                "concepts": {
                    "type": "array",
                    "objects": CONCEPT_SCHEMA
                },
                "region": {
                    "type": "object",
                    "required": ["bbox"],
                    "additionalProperties": False,
                    "properties": {
                        "bbox": {
                            "type": "object",
                            "required": ["xmin", "xmax", "ymin", "ymax"],
                            "additionalProperties": False,
                            "properties": {
                                "xmin": {"type": "number", "minimum": 0, "maximum": 1},
                                "xmax": {"type": "number", "minimum": 0, "maximum": 1},
                                "ymin": {"type": "number", "minimum": 0, "maximum": 1},
                                "ymax": {"type": "number", "minimum": 0, "maximum": 1}
                            }
                        }

                    }
                }
            }
        }
    }

    IMAGE_SCHEMA = {
        "type": "object",
        "required": [
            "data"
        ],
        "additionalProperties": False,
        "properties": {
            "id": {
                "type": ["number", "null"]
            },
            "data": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["file"],
                    "additionalProperties": False,
                    "properties": {
                        "file": {
                            "type": "string"
                        }
                    }
                }
            },
            "metadata": {
                "type": "string"
            },
            "annotations": ANNOTATION_SCHEMA,
            "splits": {
                "type": "array",
                "items": {
                    "anyOf": [
                        {"required": ["train"]},
                        {"required": ["val"]}
                    ]
                }
            }
        }
    }


def is_valid_json_with_schema(json_data, json_schema):
    """Validate a JSON using a schema"""
    try:
        validate(instance=json_data, schema=json_schema)
        return True
    except Exception:
        return False


def validate_json(json_data):
    """
    Validate a JSON using the Studio and Vulcan schema
    Returns:
    - is_valid: True if the JSON is valid
    - error: ValidationError raised if not valid
    - schema_type: Studio or Vulcan, or None if both schema raise an error at the root of the JSON
    """
    is_valid = False
    error = None
    schema_type = None
    schema_dict = {JSONSchemaType.STUDIO_HEADER: JSONSchemaType.HEADER_SCHEMA,
                   JSONSchemaType.STUDIO_INPUT: JSONSchemaType.IMAGE_SCHEMA,
                   JSONSchemaType.VULCAN: JSONSchemaType.VULCAN_JSON_SCHEMA}
    for schema_name, json_schema in schema_dict.items():
        try:
            validate(instance=json_data, schema=json_schema)
            is_valid = True
            schema_type = schema_name
            break
        except ValidationError as e:
            # If the error did not happen at the root, return the error and the current schema type (if known)
            error = e
            # If the error did not happen at the root we know the schema type of the JSON
            if len(e.absolute_path) > 0:
                schema_type = schema_name
                break
    return is_valid, error, schema_type



from deepomatic.cli.cmds.utils import Command
from ..utils import InferManager
from deepomatic.cli.cmds.utils import setup_model_cmd_line_parser


class NoopCommand(Command):
    """
        Does nothing but reading the input and outputting it in the specified format, without predictions.
        Typical usage is: deepo platform model noop -i 0 -o window
    """

    def setup(self, subparsers):
        parser = super(NoopCommand, self).setup(subparsers)
        setup_model_cmd_line_parser("platform", "noop", parser)
        return parser

    def run(self, **kwargs):
        return InferManager().input_loop(kwargs)



import os
import logging
from contextlib import contextmanager
from deepomatic.api.http_helper import HTTPHelper
from ..version import __title__, __version__


###############################################################################


LOGGER = logging.getLogger(__name__)

DEFAULT_USER_AGENT_PREFIX = '{}/{}'.format(__title__, __version__)

CLIENT = None

###############################################################################


class Client(HTTPHelper):
    def __init__(self, **kwargs):
        super(Client, self).__init__(user_agent_prefix=DEFAULT_USER_AGENT_PREFIX,
                                     version=None, **kwargs)

    def _setup_host(self, host, verify_ssl):
        if host is None:
            host = os.environ['CAMERA_SERVER_URL']
        if verify_ssl is None:
            verify_ssl = os.getenv('CAMERA_SERVER_VERIFY_TLS', '1') == '1'

        if not host.endswith('/'):
            host += '/'

        self.verify_ssl = verify_ssl
        self.host = host

    def _setup_credentials(self, app_id, api_key):
        # No credentials
        pass

    def default_headers(self):
        return {
            'User-Agent': self.user_agent,
        }


class CameraCtrl(object):
    def __init__(self, client):
        self.client = client

    def add(self, name, address):
        return self.client.post('/cameras', json={
            "name": name,
            "address": address,
        })

    def delete(self, name):
        return self.client.delete('/cameras/{}'.format(name))

    def start(self, name):
        return self.client.post('/cameras/{}/start'.format(name))

    def stop(self, name):
        return self.client.post('/cameras/{}/stop'.format(name))

    def list(self):
        return self.client.get('/cameras')


def get_client():
    global CLIENT
    if CLIENT is None:
        CLIENT = Client()
    return CLIENT


def get_camera_ctrl():
    return CameraCtrl(get_client())


@contextmanager
def start_stop_camera(name):
    LOGGER.info("Starting camera '{}'".format(name))
    get_camera_ctrl().start(name)
    try:
        yield
    except Exception:
        LOGGER.info("Stopping camera '{}'".format(name))
        get_camera_ctrl().stop(name)
        raise


@contextmanager
def temp_camera(name, address):
    LOGGER.info("Adding camera '{}'".format(name))
    camera = get_camera_ctrl().add(name, address)
    name = camera['name']

    try:
        with start_stop_camera(name):
            yield camera
    except Exception:
        LOGGER.info("Deleting camera '{}'".format(name))
        get_camera_ctrl().delete(name)
        raise


def add_start_camera(name, address):
    LOGGER.info("Adding camera '{}'".format(name))
    get_camera_ctrl().add(name, address)
    LOGGER.info("Starting camera '{}'".format(name))
    get_camera_ctrl().start(name)


def add(args):
    get_camera_ctrl().add(args.camera_name, args.camera_address)
    LOGGER.info("Camera '{}' added".format(args.camera_name))


def start(args):
    get_camera_ctrl().start(args.camera_name)
    LOGGER.info("Camera '{}' started".format(args.camera_name))


def stop(args):
    get_camera_ctrl().stop(args.camera_name)
    LOGGER.info("Camera '{}' stopped".format(args.camera_name))


def delete(args):
    get_camera_ctrl().delete(args.camera_name)
    LOGGER.info("Camera '{}' delete".format(args.camera_name))


def list(args):
    cameras = get_camera_ctrl().list(args.camera_name)
    LOGGER.info("Cameras: \n{}".format(cameras))


def stream(args):
    add_start_camera(args.camera_name, args.camera_address)



class DeepoCLIException(Exception):
    pass


class DeepoRPCUnavailableError(DeepoCLIException):
    pass


class DeepoRPCRecognitionError(DeepoCLIException):
    pass


class DeepoCLICredentialsError(DeepoCLIException):
    pass


class DeepoWorkflowError(DeepoCLIException):
    pass


class DeepoUnknownOutputError(DeepoCLIException):
    pass


class DeepoSaveJsonToFileError(DeepoCLIException):
    pass


class DeepoOpenJsonError(DeepoCLIException):
    pass


class DeepoFPSError(DeepoCLIException):
    pass


class DeepoVideoOpenError(DeepoCLIException):
    pass


class DeepoInputError(DeepoCLIException):
    pass


class DeepoPredictionJsonError(DeepoCLIException):
    pass


class DeepoUploadJsonError(DeepoCLIException):
    pass


# Inference errors


class InferenceError(DeepoCLIException):
    def __init__(self, error):
        super(InferenceError, self).__init__(str(error))
        self.error = error


class SendInferenceError(InferenceError):
    pass


class ResultInferenceError(InferenceError):
    pass


class ResultInferenceTimeout(ResultInferenceError):
    def __init__(self, timeout=None):
        self.timeout = timeout
        error = 'timeout reached'
        if timeout is not None:
            error += ' after {}'.format(timeout)
        super(ResultInferenceTimeout, self).__init__(error)



from .utils import _SiteCommand, SiteManager


class InstallCommand(_SiteCommand):
    """
        Install a site
    """

    def run(self, site_id, **kwargs):
        app_id = SiteManager().install(site_id)
        if app_id is None:
            print('Site', site_id, 'already installed')
        else:
            print('Site', site_id, 'installed with app', app_id)



from deepomatic.cli.common import SUPPORTED_IMAGE_INPUT_FORMAT
from deepomatic.cli.cmds import parser_helpers
from deepomatic.cli.cmds.utils import Command
from deepomatic.cli.cmds.platform.utils import AddImageManager


class AddImagesCommand(Command):
    """
        Upload simple images to Deepomatic Platform.
        Typical usage is: deepo platform add_images -i img.png -p myproject -o myorg
    """

    def setup(self, subparsers):
        parser = super(AddImagesCommand, self).setup(subparsers)

        # Define studio group for add_images
        group = parser.add_argument_group('studio arguments')
        help_msg = "Deepomatic Studio project name. Note: the argument --dataset was changed to --project"
        group.add_argument('-p', '--project', required=True, help=help_msg, type=str)
        group.add_argument('-o', '--org', required=True, help="Deepomatic Studio org slug.", type=str)

        input_group = parser_helpers.add_common_cmd_group(parser, 'input')
        # Define input group for add_images
        input_group.add_argument('-i', '--input', type=str, nargs='+', required=True,
                                 help="One or several input path, either an image file (*{}),"
                                 " a directory, or a Studio txt (*.txt).".format(
                                     ', *'.join(SUPPORTED_IMAGE_INPUT_FORMAT)
                                 ))
        input_group.add_argument('--txt', dest='txt_file', action='store_true',
                                 help='Look for txt files instead of images.')
        parser_helpers.add_recursive_argument(input_group)

        parser.add_argument('--set_metadata_path', dest='set_metadata_path',
                            action='store_true',
                            help='Add the relative path as metadata.')
        return parser

    def run(self, **kwargs):
        return AddImageManager().upload(kwargs)



from ...utils import PlatformCommand
from ..utils import DrivePlatformManager


class CreateCommand(PlatformCommand):
    """Create a new DriveApp."""

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-n', '--name', required=True, type=str, help="DriveApp name")
        parser.add_argument('-d', '--description', type=str, help="DriveApp description")
        parser.add_argument('-s', '--services', required=True, nargs="*", type=str,
                            help="Services needed by the DriveApp.", default=[])
        return parser

    def run(self, name, description, services, **kwargs):
        return DrivePlatformManager().create_app(name, description, services)



from ..utils import _CurrentSiteCommand, SiteManager


class StartCommand(_CurrentSiteCommand):
    """
        Start the site
    """

    def run(self, **kwargs):
        manager = SiteManager()
        print('Starting site', manager.current())
        manager.start()
        print('Site started')



from ..utils import _CurrentSiteCommand, SiteManager


class LogsCommand(_CurrentSiteCommand):
    """
        Get the logs of the running site
    """

    def run(self, **kwargs):
        SiteManager().logs()



from .utils import _SiteCommand, SiteManager


class UninstallCommand(_SiteCommand):
    """
        Uninstall a site
    """

    def run(self, site_id, **kwargs):
        SiteManager().uninstall(site_id)
        # TODO: feedback



# PYTHON_ARGCOMPLETE_OK
import sys
from .cli_parser import run


def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]
    run(args)


if __name__ == "__main__":
    main()



from ...utils import PlatformCommand
from ..utils import DrivePlatformManager


class DeleteCommand(PlatformCommand):
    """Delete a DriveApp."""

    def setup(self, subparsers):
        parser = super(DeleteCommand, self).setup(subparsers)
        parser.add_argument('-i', '--drive_app_id', required=True, type=str, help="DriveApp id")
        return parser

    def run(self, drive_app_id, **kwargs):
        return DrivePlatformManager().delete_app(drive_app_id)



# -*- coding: utf-8 -*-

import os
import sys
import json
import cv2
import logging
import traceback
from .thread_base import Thread
from .common import (Empty, write_frame_to_disk, SUPPORTED_IMAGE_OUTPUT_FORMAT,
                     SUPPORTED_VIDEO_OUTPUT_FORMAT, SUPPORTED_FOURCC, BGR_TO_COLOR_SPACE)
from .cmds.studio_helpers.vulcan2studio import transform_json_from_vulcan_to_studio
from .exceptions import DeepoUnknownOutputError, DeepoSaveJsonToFileError


LOGGER = logging.getLogger(__name__)
DEFAULT_OUTPUT_FPS = 25


try:
    # https://stackoverflow.com/questions/908331/how-to-write-binary-data-to-stdout-in-python-3
    write_bytes_to_stdout = sys.stdout.buffer.write
except AttributeError:
    write_bytes_to_stdout = sys.stdout.write


def save_json_to_file(json_data, json_path):
    try:
        with open('%s.json' % json_path, 'w') as f:
            LOGGER.debug('Writing %s.json' % json_path)
            json.dump(json_data, f)
            LOGGER.debug('Writing %s.json done' % json_path)
    except Exception:
        raise DeepoSaveJsonToFileError("Could not save file {} in json format: {}".format(json_path, traceback.format_exc()))


def get_output(descriptor, kwargs):
    if descriptor is not None:
        if ImageOutputData.is_valid(descriptor):
            return ImageOutputData(descriptor, **kwargs)
        elif VideoOutputData.is_valid(descriptor):
            return VideoOutputData(descriptor, **kwargs)
        elif JsonOutputData.is_valid(descriptor):
            return JsonOutputData(descriptor, **kwargs)
        elif DirectoryOutputData.is_valid(descriptor):
            return DirectoryOutputData(descriptor, **kwargs)
        elif descriptor == 'stdout':
            return StdOutputData(**kwargs)
        elif descriptor == 'window':
            return DisplayOutputData(**kwargs)
        else:
            raise DeepoUnknownOutputError("Unknown output '{}'".format(descriptor))
    else:
        return DisplayOutputData(**kwargs)


def get_outputs(descriptors, kwargs):
    return [get_output(descriptor, kwargs) for descriptor in descriptors]


class NotProcessedYet(object):
    # OutputThread can receive frames in wrong order
    # When a frame arrives and is not the one we want to process,
    # we cache it internally and return a NotProcessedYet object
    # Then we override OutputThread.task_done so that we don't consider
    # the task to be done when NotProcessedYet is returned
    pass


class OutputThread(Thread):
    NOT_PROCESSED_YET = NotProcessedYet()

    def __init__(self, exit_event, input_queue, output_queue, current_messages,
                 on_progress, postprocessing, **kwargs):
        super(OutputThread, self).__init__(exit_event, input_queue,
                                           output_queue, current_messages)
        self.args = kwargs
        # Opencv images are BGR by default, no need to convert
        output_color_space_str = kwargs.get('output_color_space')
        if output_color_space_str is None or output_color_space_str == 'BGR':
            self.output_color_space = None
        else:
            self.output_color_space = BGR_TO_COLOR_SPACE[output_color_space_str]

        self.on_progress = on_progress
        self.postprocessing = postprocessing
        self.frames_to_check_first = {}
        self.frame_to_output = None
        # Update output fps to default value if none was specified.
        # Logs information only if one of the outputs uses fps.
        if not kwargs['output_fps']:
            kwargs['output_fps'] = DEFAULT_OUTPUT_FPS
            self.outputs = get_outputs(self.args.get('outputs', None), self.args)
            for output in self.outputs:
                if isinstance(output, VideoOutputData) or isinstance(output, DisplayOutputData):
                    LOGGER.info('No --output_fps value specified for output, using default value of {}.'.format(DEFAULT_OUTPUT_FPS))
                    break
        else:
            self.outputs = get_outputs(self.args.get('outputs', None), self.args)

    def close(self):
        self.frames_to_check_first = {}
        self.frame_to_output = None
        for output in self.outputs:
            output.close()

    def can_stop(self):
        return super(OutputThread, self).can_stop() and \
            len(self.frames_to_check_first) == 0

    def pop_input(self):
        # looking into frames we popped earlier
        if self.frame_to_output is None:
            self.frame_to_output = self.current_messages.pop_oldest()
            if self.frame_to_output is None:
                raise Empty()

        frame = self.frames_to_check_first.pop(self.frame_to_output, None)
        if frame is None:
            frame = super(OutputThread, self).pop_input()
        return frame

    def put_to_output(self, frame_out):
        if frame_out is self.NOT_PROCESSED_YET:
            # Nothing to output, the frame has not been processed yet
            return
        super(OutputThread, self).put_to_output(frame_out)

    def task_done(self, frame_in, frame_out):
        if frame_out is self.NOT_PROCESSED_YET:
            # We kept the frame for later, the task is not done
            return

        super(OutputThread, self).task_done(frame_in, frame_out)
        self.frame_to_output = None

        if self.output_queue is None:
            # process_msg() returns frame None when output_queue is None
            assert frame_out is None
            # Reporting success only if last pool of the pipeline
            self.current_messages.report_success()

    def process_msg(self, frame):
        if self.frame_to_output != frame.frame_number:
            self.frames_to_check_first[frame.frame_number] = frame
            # We keep it for later
            return self.NOT_PROCESSED_YET

        if self.postprocessing is not None:
            self.postprocessing(frame)
        else:
            frame.output_image = frame.image  # we output the original image

        # Opencv images are BGR by default
        if self.output_color_space is not None and frame.output_image is not None:
            frame.output_image = cv2.cvtColor(frame.output_image, self.output_color_space)

        for output in self.outputs:
            output.output_frame(frame)

        if self.on_progress:
            self.on_progress()

        if self.output_queue is not None:
            # In some case we might attach other pools after the output pool
            # In this case we return the frame for the next processing
            return frame
        # No other pool attached
        # End of pipeline
        return None


class OutputData(object):
    def __init__(self, descriptor, **kwargs):
        self._descriptor = descriptor
        self._args = kwargs

    def close(self):
        pass

    def output_frame(self, frame):
        # override this to output the results of the frame
        raise NotImplementedError()


class ImageOutputData(OutputData):
    @classmethod
    def is_valid(cls, descriptor):
        _, ext = os.path.splitext(descriptor)
        return ext.lower() in SUPPORTED_IMAGE_OUTPUT_FORMAT

    def __init__(self, descriptor, **kwargs):
        super(ImageOutputData, self).__init__(descriptor, **kwargs)
        self._i = 0

    def output_frame(self, frame):
        self._i += 1
        path = self._descriptor
        try:
            path = path % self._i
        except TypeError:
            pass
        finally:
            write_frame_to_disk(frame, path)


class VideoOutputData(OutputData):
    @classmethod
    def is_valid(cls, descriptor):
        _, ext = os.path.splitext(descriptor)
        return ext.lower() in SUPPORTED_VIDEO_OUTPUT_FORMAT

    def __init__(self, descriptor, **kwargs):
        super(VideoOutputData, self).__init__(descriptor, **kwargs)
        ext = os.path.splitext(descriptor)[1].lower()
        if ext not in SUPPORTED_FOURCC:
            raise Exception("Unsupported video output extension {}".format(ext))

        FOURCC = SUPPORTED_FOURCC[ext]
        # if fourcc is provided, try to use it
        if kwargs.get('fourcc') is not None:
            fourcc = kwargs['fourcc']
        else:
            # default to the first supported fourcc
            fourcc = FOURCC[0]

        if fourcc not in FOURCC:
            LOGGER.warning("The specified fourcc {} might not be muxable into the container format {}".format(fourcc, ext))

        self._fourcc = cv2.VideoWriter_fourcc(*fourcc)
        self._fps = kwargs['output_fps']
        self._writer = None

    def close(self):
        if self._writer is not None:
            self._writer.release()
        self._writer = None

    def output_frame(self, frame):
        if frame.output_image is None:
            LOGGER.warning('No frame to output.')
        else:
            if self._writer is None:
                LOGGER.debug('Writing %s' % self._descriptor)
                self._writer = cv2.VideoWriter(self._descriptor, self._fourcc,
                                               self._fps, (frame.output_image.shape[1],
                                                           frame.output_image.shape[0]))
            self._writer.write(frame.output_image)


class StdOutputData(OutputData):
    def __init__(self, **kwargs):
        super(StdOutputData, self).__init__(None, **kwargs)

    def output_frame(self, frame):
        if frame.output_image is None:
            print(json.dumps(frame.predictions))
        else:
            write_bytes_to_stdout(frame.output_image.tobytes())


class DisplayOutputData(OutputData):
    def __init__(self, **kwargs):
        super(DisplayOutputData, self).__init__(None, **kwargs)
        self._fps = kwargs['output_fps']
        self._window_name = 'Deepomatic'
        self._fullscreen = kwargs.get('fullscreen', False)

        if self._fullscreen:
            cv2.namedWindow(self._window_name, cv2.WINDOW_NORMAL)
            cv2.setWindowProperty(self._window_name,
                                  cv2.WND_PROP_FULLSCREEN,
                                  cv2.WINDOW_FULLSCREEN)

    def output_frame(self, frame):
        if frame.output_image is None:
            LOGGER.warning('No frame to output.')
        else:
            cv2.imshow(self._window_name, frame.output_image)
            try:
                ms = 1000 // int(self._fps)
            except Exception:
                ms = 1
            if cv2.waitKey(ms) & 0xFF == ord('q'):
                cv2.destroyAllWindows()
                cv2.waitKey(1)
                sys.exit()

    def close(self):
        if cv2.waitKey(0) & 0xFF == ord('q'):
            cv2.destroyAllWindows()
            cv2.waitKey(1)


class WildCardType():
    # Used to differentiate the different type of output json
    NONE = None             # pred.json, not completed
    INTEGER = 'integer'     # pred_%04d.json, completed with frame number
    STRING = 'string'       # pred_%s.json, completed with frame name


class JsonOutputData(OutputData):
    @classmethod
    def is_valid(cls, descriptor):
        _, ext = os.path.splitext(descriptor)
        return ext.lower() == '.json'

    def __init__(self, descriptor, **kwargs):
        super(JsonOutputData, self).__init__(descriptor, **kwargs)
        self._i = 0
        self._to_studio_format = kwargs.get('studio_format')
        self._preserve_input_dir_structure = False
        self._input_path = None

        # Check if the output is a string wildcard
        try:
            descriptor % 'string'
            self._wildcard_type = WildCardType.STRING
            self._all_predictions = None
            if kwargs.get('recursive', False):
                self._preserve_input_dir_structure = True
                self._input_path = kwargs.get('input')
        except TypeError:
            # Check if the output is an integer wildcard
            try:
                descriptor % 0
                self._wildcard_type = WildCardType.INTEGER
                self._all_predictions = None
            # Otherwise it's a pure json
            except TypeError:
                self._wildcard_type = WildCardType.NONE
                if self._to_studio_format:
                    self._all_predictions = {'tags': [], 'images': []}
                else:
                    self._all_predictions = []

    def close(self):
        if self._all_predictions is not None:
            json_path = os.path.splitext(self._descriptor)[0]
            save_json_to_file(self._all_predictions, json_path)

    def output_frame(self, frame):
        self._i += 1
        if frame.predictions is None:
            # For noop command
            LOGGER.warning('No predictions to output.')
            return
        predictions = frame.predictions
        predictions['location'] = frame.filename
        predictions['data'] = {
            'framename': frame.name,
            'original_filename': frame.filename
        }
        if self._to_studio_format:
            predictions = transform_json_from_vulcan_to_studio(predictions)

        if self._all_predictions is not None:
            # If the json is not a wildcard we store prediction to write then to file a the end in close()
            if self._to_studio_format:
                self._all_predictions['images'] += predictions['images']
                self._all_predictions['tags'] = list(
                    set(self._all_predictions['tags'] + predictions['tags'])
                )
            else:
                self._all_predictions.append(predictions)
        # Otherwise we write them to file directly
        else:
            # Build the prediction json path
            if self._wildcard_type == WildCardType.INTEGER:
                json_path = os.path.splitext(self._descriptor % self._i)[0]
            elif self._wildcard_type == WildCardType.STRING:
                if not self._preserve_input_dir_structure:
                    json_path = os.path.splitext(self._descriptor % frame.name)[0]
                else:
                    # Build the output directory with input structure
                    output_base_dir = os.path.dirname(self._descriptor)
                    input_rel_dir = os.path.dirname(os.path.relpath(frame.filename, self._input_path))
                    output_full_dir = os.path.join(output_base_dir, input_rel_dir)

                    # Build the final path
                    json_file = os.path.splitext(os.path.basename(self._descriptor) % frame.name)[0]
                    json_path = os.path.join(output_full_dir, json_file)

                    # Build the json directory if it doesn't exist
                    if output_full_dir and not os.path.isdir(output_full_dir):
                        os.makedirs(output_full_dir)
            save_json_to_file(predictions, json_path)


class DirectoryOutputData(OutputData):
    @classmethod
    def is_valid(cls, descriptor):
        return (os.path.exists(descriptor) and os.path.isdir(descriptor))

    def __init__(self, descriptor, **kwargs):
        super(DirectoryOutputData, self).__init__(descriptor, **kwargs)
        self._input = self._args['input']

    def output_frame(self, frame):
        # If the input is a directory, then preserve directory structure, see below
        # - Command: deepo -i dir1/ -R -o dir2/subdir2/ -r 123...
        # - Input directory structure:
        #     dir1
        #     ├── subdir1
        #     │   ├── img1.jpg
        #     │   └── img2.jpg
        #     └── video.mp4
        # - Output directory structure:
        #     dir2
        #     └── subdir2
        #         ├── subdir1
        #         │   ├── img1_123.jpg
        #         │   └── img2_123.jpg
        #         ├── video_00001_123.jpg
        #         ├── ...
        #         └── video_xxxxx_123.jpg
        # Otherwise implement a flat directory structure
        if os.path.isdir(self._input):
            rel_path = os.path.relpath(frame.filename, self._input)
            rel_dir = os.path.dirname(rel_path)
            root_dir = os.path.join(self._descriptor, rel_dir)
        else:
            root_dir = self._descriptor
        if not os.path.isdir(root_dir):
            os.makedirs(root_dir)

        # If the input is an image, then use the same extension if supported
        _, ext = os.path.splitext(frame.filename)
        if ext.lower() not in SUPPORTED_IMAGE_OUTPUT_FORMAT:
            # Otherwise defaults to jpg
            ext = '.jpg'

        # Finally write the image to file with its name
        path = os.path.join(root_dir, "{}{}".format(frame.name, ext))
        write_frame_to_disk(frame, path)



# Common parsers functions used by various commands
COMMON_GROUPS = {
    'input', 'output'
}


def add_common_cmd_group(parser, label):
    assert label in COMMON_GROUPS
    return parser.add_argument_group('{} arguments'.format(label.lower()))


def add_recursive_argument(parser):
    parser.add_argument('-R', '--recursive', dest='recursive', action='store_true',
                        help='If a directory input is used, goes through all files in subdirectories.')


def add_verbose_argument(parser):
    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        help='Increase output verbosity.')



from ..utils import _CurrentSiteCommand, SiteManager


class StatusCommand(_CurrentSiteCommand):
    """
        Get the site status
    """

    def run(self, **kwargs):
        SiteManager().status()



from .utils import _CurrentSiteCommand, SiteManager


class ShowCommand(_CurrentSiteCommand):
    """
        Show the site
    """

    def run(self, **kwargs):
        print('site:', SiteManager().current())
        print('app', SiteManager().current_app())



from ..utils import _CurrentSiteCommand, SiteManager


class StopCommand(_CurrentSiteCommand):
    """
        Stop the site
    """

    def run(self, **kwargs):
        SiteManager().stop()
        # TODO: feedback



import logging
from .workflow_abstraction import AbstractWorkflow
from ..exceptions import (SendInferenceError,
                          ResultInferenceError,
                          ResultInferenceTimeout)
from .. import exceptions
import deepomatic.api.client
import deepomatic.api.inputs
from ..version import __title__, __version__
from tenacity import stop_never
from deepomatic.api.exceptions import TaskError, TaskTimeout, BadStatus, DeepomaticException

LOGGER = logging.getLogger(__name__)


class CloudRecognition(AbstractWorkflow):

    class InferResult(AbstractWorkflow.AbstractInferResult):
        def __init__(self, task):
            self._task = task

        def get_predictions(self, timeout):
            try:
                self._task.wait(timeout=timeout)
                return self._task['data']
            except BadStatus as e:
                # HTTP Error
                raise ResultInferenceError(e)
            except TaskError:
                # Task is in error
                raise ResultInferenceError(self._task.data())
            except TaskTimeout:
                # Task did not finish on time
                raise ResultInferenceTimeout(timeout)

        def __str__(self):
            return '<{} task_id={}>'.format(self.__class__.__qualname__, self._task.pk)

    def close(self):
        self._client.http_helper.session.close()

    def __init__(self, recognition_version_id):
        super(CloudRecognition, self).__init__('r{}'.format(recognition_version_id))
        self._id = recognition_version_id

        # Retry indefinitely until the API is available
        http_retry = deepomatic.api.http_retry.HTTPRetry(stop=stop_never)

        user_agent_prefix = '{}/{}'.format(__title__, __version__)
        try:
            self._client = deepomatic.api.client.Client(user_agent_prefix=user_agent_prefix,
                                                        http_retry=http_retry)
        except DeepomaticException:  # TODO later replace with CredentialsNotFound
            error = ('Credentials not found.'
                     ' Please define the DEEPOMATIC_API_KEY environment variable to use cloud-based recognition models.')
            raise exceptions.DeepoCLICredentialsError(error)

        self._model = None
        try:
            recognition_version_id = int(recognition_version_id)
        except ValueError:
            LOGGER.warning("Cannot cast recognition ID into a number, trying with a public recognition model")
            self._model = self._client.RecognitionSpec.retrieve(recognition_version_id)
        if self._model is None:
            self._model = self._client.RecognitionVersion.retrieve(recognition_version_id)

    def infer(self, encoded_image_bytes, _useless_push_client, _useless_frame_name):
        # _useless_push_client and _useless_frame_name are used for the rpc and json workflows
        try:
            return self.InferResult(self._model.inference(
                inputs=[deepomatic.api.inputs.ImageInput(encoded_image_bytes, encoding="binary")],
                show_discarded=True,
                return_task=True,
                wait_task=False))
        except BadStatus as e:
            # HTTP error
            raise SendInferenceError(e)



import os
import logging

try:
    from builtins import FileExistsError
except ImportError:
    FileExistsError = OSError

from deepomatic.api.http_helper import HTTPHelper
from deepomatic.cli.cmds.utils import CommandResult

from .add_images import DEFAULT_USER_AGENT_PREFIX


LOGGER = logging.getLogger(__name__)


class DrivePlatformManager(object):
    def __init__(self, client_cls=HTTPHelper):
        self.drive_client = client_cls()

    def create_app(self, name, description, services):
        data_app = {
            'name': name,
            'desc': description,
            'services': [{"name": service} for service in services]
        }
        ret = self.drive_client.post('/apps', data=data_app)
        return CommandResult("created", "drive_app", ret)

    def update_app(self, app_id, name, description):
        data = {}

        if name is not None:
            data['name'] = name

        if description is not None:
            data['desc'] = description

        ret = self.drive_client.patch('/apps/{}'.format(app_id), data=data)
        return CommandResult("updated", "drive_app", ret)

    def delete_app(self, app_id):
        self.drive_client.delete('/apps/{}'.format(app_id))
        return CommandResult("deleted", "drive_app", {"id": app_id})

    def create_app_version(self, app_id, name, description, app_specs, version_ids):
        data = {
            'app_id': app_id,
            'name': name,
            'app_specs': app_specs,
            'recognition_version_ids': version_ids,
            # FIXME: To update when endpoint are updated.
            'resources': []
        }
        if description is not None:
            data['desc'] = description

        ret = self.drive_client.post('/app-versions', data=data)
        return CommandResult("created", "drive_app_version", ret)

    def update_app_version(self, app_version_id, name, description):
        data = {}

        if name is not None:
            data['name'] = name

        if description is not None:
            data['desc'] = description

        ret = self.drive_client.patch('/app-versions/{}'.format(app_version_id), data=data)
        return CommandResult("updated", "drive_app_version", ret)

    def delete_app_version(self, app_version_id):
        self.drive_client.delete('/app-versions/{}'.format(app_version_id))
        return CommandResult("deleted", "drive_app_version", {"id": app_version_id})


class EngagePlatformManager(object):
    def __init__(self, client_cls=HTTPHelper):
        try:
            ENGAGE_API_URL = os.environ['ENGAGE_API_URL']
        except KeyError as e:
            raise SystemExit(e, "environment variable ENGAGE_API_URL is missing.")

        try:
            slug = os.environ['ORGANIZATION_SLUG']
            FS_URL_PREFIX = "engage/fs/on-site/orgs/{}".format(slug)
        except KeyError as e:
            raise SystemExit(e, "environment variable ORGANIZATION_SLUG is missing.")

        user_agent_prefix = DEFAULT_USER_AGENT_PREFIX
        self.engage_client = client_cls(
            host=ENGAGE_API_URL,
            user_agent_prefix=user_agent_prefix,
            version=""
        )

        self.engage_app_endpoint = "{}/apps".format(FS_URL_PREFIX)
        self.version_create_from_endpoint = FS_URL_PREFIX + "/app-versions/{}/create-from"

    def create_app(self, name, application_type):
        data = {"name": name}

        if application_type:
            data["application_type"] = application_type

        response = self.engage_client.post(
            '{}'.format(self.engage_app_endpoint),
            data=data
        )

        return CommandResult("created", "engage_app", response)

    def delete_app(self, app_id):
        self.engage_client.delete('{}/{}'.format(self.engage_app_endpoint, app_id))
        return CommandResult("deleted", "engage_app", {"id": app_id})

    def create_app_version(self,
                           app_id,
                           workflow_path,
                           custom_nodes_path,
                           recognition_version_ids,
                           base_major_version):

        data = {'recognition_version_ids': recognition_version_ids}

        if base_major_version:
            data['base_major_version'] = base_major_version

        try:
            files = {'workflow_yaml': open(workflow_path, 'r')}
            if custom_nodes_path:
                files['custom_nodes_py'] = open(custom_nodes_path, 'r')

            response = self.engage_client.post(
                '{}/{}/versions'.format(self.engage_app_endpoint, app_id),
                data=data,
                files=files,
                content_type='multipart/mixed'
            )
        finally:
            for file in files.values():
                file.close()

        return CommandResult(
            "created",
            "engage_app_version",
            response,
            ["id", "drive_app_version_id"],
            "version={}.{}".format(response["major"], response["minor"])
        )

    def create_app_version_from(self,
                                origin,
                                base_major_version,
                                workflow_path,
                                custom_nodes_path,
                                recognition_version_ids):
        data = {}
        files = {}
        kwargs = {}

        if recognition_version_ids:
            data['recognition_version_ids'] = recognition_version_ids
        if base_major_version:
            data['base_major_version'] = base_major_version
        kwargs['data'] = data

        try:
            if workflow_path:
                files['workflow_yaml'] = open(workflow_path, 'r')
            if custom_nodes_path:
                files['custom_nodes_py'] = open(custom_nodes_path, 'r')
            if files:
                kwargs['files'] = files
                kwargs['content_type'] = "multipart/mixed"

            response = self.engage_client.post(
                self.version_create_from_endpoint.format(origin),
                **kwargs
            )
        finally:
            for file in files.values():
                file.close()

        return CommandResult(
            "created",
            "engage_app_version",
            response,
            ["id", "drive_app_version_id"],
            "version={}.{} from={}".format(response["major"], response["minor"], origin)
        )



from ..utils import Command
from ...lib.site import SiteManager


class ManifestCommand(Command):
    """
        Get a deployment manifest
    """

    def setup(self, subparsers):
        parser = super(ManifestCommand, self).setup(subparsers)
        parser.add_argument('-i', '--id', required=True, type=str, help="Site id")
        parser.add_argument('-t', '--target', required=True, type=str, help="deployment target")
        return parser

    def run(self, id, target, **kwargs):
        return SiteManager().get_deployment_manifest(id, target)



from ...utils import Command


class CurrentCommand(Command):
    """
        Control the current site
    """

    from .start import StartCommand
    from .stop import StopCommand
    from .logs import LogsCommand
    from .status import StatusCommand

    from .camera import CameraCommand



import logging
from .cloud_workflow import CloudRecognition
from .rpc_workflow import RpcRecognition
from .json_workflow import JsonRecognition
from ..exceptions import DeepoWorkflowError


LOGGER = logging.getLogger(__name__)


def get_workflow(args):
    # Retrieve recognition arguments
    recognition_id = args.get('recognition_id')
    amqp_url = args.get('amqp_url')
    routing_key = args.get('routing_key')
    pred_from_file = args.get('pred_from_file')
    model_command = args.get('model_command')

    # Check whether we should use predictions from a json, rpc deployment or the cloud API
    if pred_from_file:
        LOGGER.debug('Using JSON workflow with recognition_id {}'.format(recognition_id))
        return JsonRecognition(recognition_id, pred_from_file)
    elif all([amqp_url, routing_key]):
        LOGGER.debug('Using RPC workflow with'
                     ' recognition_id {}, amqp_url {} and routing_key {}'.format(recognition_id, amqp_url, routing_key))
        return RpcRecognition(recognition_id, amqp_url, routing_key)
    elif recognition_id:
        LOGGER.debug('Using Cloud workflow with recognition_id {}'.format(recognition_id))
        return CloudRecognition(recognition_id)
    elif model_command == 'noop':
        return None
    else:
        raise DeepoWorkflowError("Couldn't get workflow based on args {}".format(args))


__all__ = ["get_workflow"]



from ..utils import Command
from .utils import SiteManager


class ListCommand(Command):
    """
        Get the list of installed sites
    """

    def run(self, **kwargs):
        print('\n'.join(SiteManager().list()))



from ...utils import (
    PlatformCommand,
    valid_json
)
from ..utils import DrivePlatformManager


class CreateCommand(PlatformCommand):
    """Create a new DriveAppVersion."""

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-n', '--name', required=True, type=str, help="DriveAppVersion name")
        parser.add_argument('-d', '--description', type=str, help="DriveAppVersion description")
        parser.add_argument('-i', '--drive_app_id', required=True, type=str, help="DriveApp id for this DriveAppVersion")
        parser.add_argument('-s', '--app_specs', default=None, required=True,
                            type=valid_json, help="""
                            JSON specs for the app (if workflow not provided).
                            Example:
                            '[{"recognition_spec_id": 123, "queue_name": "spec_123.forward"}]'
                            """)
        parser.add_argument('-r', '--recognition-version-ids', required=True, nargs="*", type=int,
                            help="List of Recognition Version Id, one for each Recognition Spec in the App", default=[])
        return parser

    def run(self, drive_app_id, name, description, app_specs, recognition_version_ids, **kwargs):
        return DrivePlatformManager().create_app_version(
            drive_app_id,
            name,
            description,
            app_specs,
            recognition_version_ids
        )



# -*- coding: utf-8 -*-
import os
import json
import uuid
import logging
from ...thread_base import Greenlet
from ...common import SUPPORTED_IMAGE_INPUT_FORMAT
from ...json_schema import JSONSchemaType, validate_json


BATCH_SIZE = int(os.getenv('DEEPOMATIC_CLI_ADD_IMAGES_BATCH_SIZE', '10'))
LOGGER = logging.getLogger(__name__)


class UploadImageGreenlet(Greenlet):
    def __init__(self, exit_event, input_queue, current_messages,
                 helper, task, on_progress=None, set_metadata_path=False,
                 **kwargs):
        super(UploadImageGreenlet, self).__init__(exit_event, input_queue,
                                                  current_messages=current_messages)
        self.args = kwargs
        self.on_progress = on_progress
        self._helper = helper
        self._task = task
        self._set_metadata_path = set_metadata_path

    def process_msg(self, msg):
        url, batch = msg
        files = {}
        meta = {}
        for file in batch:
            try:
                self.current_messages.report_message()

                # Update file
                files.update({file['key']: open(file['path'], 'rb')})

                # Update corresponding metadata
                file_meta = file.get('meta', {})
                file_metadata = file_meta.get('metadata', '{}')
                if type(file_metadata) is str:
                    file_metadata = json.loads(file_metadata)
                if self._set_metadata_path:
                    file_metadata.update({'image_path': file['path']})
                file_meta['metadata'] = json.dumps(file_metadata)
                meta[file['key']] = file_meta
            except RuntimeError as e:
                self.current_messages.report_error()
                LOGGER.error('Something when wrong with {}: {}. Skipping it.'.format(file['path'], e))
        try:
            rq = self._helper.post(url, data={"objects": json.dumps(meta)}, content_type='multipart/mixed', files=files)
            self._task.retrieve(rq['task_id'])
        except RuntimeError as e:
            self.current_messages.report_errors(len(meta))
            LOGGER.error("Failed to upload batch of images {}: {}.".format(files, e))

        for fd in files.values():
            try:
                fd.close()
            except Exception:
                pass

        if self.on_progress:
            self.on_progress(len(batch))
        self.current_messages.report_successes(len(batch))


class DatasetFiles(object):
    def __init__(self, helper, output_queue):
        self._helper = helper
        self.output_queue = output_queue

    def flush_batch(self, url, batch):
        if len(batch) > 0:
            self.output_queue.put((url, batch))
        return []

    def fill_flush_batch(self, url, batch, path, meta=None):
        image_key = uuid.uuid4().hex
        img = {"key": image_key, "path": path}
        if meta is not None:
            meta['location'] = image_key
            img['meta'] = meta
        batch.append(img)
        if len(batch) >= BATCH_SIZE:
            return self.flush_batch(url, batch)
        return batch

    def fill_queue(self, files, org_slug, project_name):
        total_files = 0
        url = 'orgs/{}/datasets/{}/images/batch/'.format(org_slug, project_name)
        batch = []

        for upload_file in files:
            extension = os.path.splitext(upload_file)[1].lower()
            # If it's an image file add it to the queue
            if extension in SUPPORTED_IMAGE_INPUT_FORMAT:
                meta = {'file_type': 'image'}
                batch = self.fill_flush_batch(url, batch, upload_file, meta=meta)
                total_files += 1

            # If it's a txt, deal with it accordingly
            elif extension == '.txt':
                with open(upload_file, 'r') as fd:
                    line_number = 0
                    for line in fd:
                        line_number += 1
                        line = json.loads(line)
                        is_valid_json, error, schema_type = validate_json(line)
                        if schema_type == JSONSchemaType.STUDIO_HEADER:
                            self.post_header(org_slug, project_name, line)
                        elif schema_type == JSONSchemaType.STUDIO_INPUT:
                            input = line.pop('data')[0]
                            if 'file' in input:
                                img_loc = input['file']
                                file_path = os.path.join(os.path.dirname(upload_file), img_loc)
                                if not os.path.isfile(file_path):
                                    LOGGER.error("Can't find file named {}".format(img_loc))
                                    continue
                                batch = self.fill_flush_batch(url, batch, file_path, meta=line)
                                total_files += 1
                            else:
                                LOGGER.error("Line {} invalid \"{}\". Skipping it".format(line_number, line))
                        else:
                            LOGGER.error("Line {} invalid \"{}\". Skipping it".format(line_number, line))
            else:
                LOGGER.info("File {} not supported. Skipping it.".format(upload_file))
        self.flush_batch(url, batch)
        return total_files

    def post_files(self, org_slug, project_name, files):
        # Retrieve endpoint
        try:
            request = 'orgs/{}/projects/{}/'.format(org_slug, project_name)
            self._helper.get(request)
        except RuntimeError:
            raise RuntimeError("Can't find the project {}".format(project_name))
        return self.fill_queue(files, org_slug, project_name)

    def post_header(self, org_slug, project_name, project_header):
        request = 'orgs/{}/projects/{}/create_views/'.format(org_slug, project_name)
        self._helper.post(request, data=project_header)



from ..utils import Command


class SiteCommand(Command):
    """
        Operations on your site (local or cloud)
        You can choose from the following commands to manage your local sites
    """

    # Offline
    from .use import UseCommand
    from .list import ListCommand
    from .load import LoadCommand
    from .save import SaveCommand
    from .uninstall import UninstallCommand
    from .rollback import RollbackCommand
    from .show import ShowCommand
    from .current import CurrentCommand

    # Using the API
    from .install import InstallCommand
    from .upgrade import UpgradeCommand
    from .create import CreateCommand
    from .update import UpdateCommand
    from .delete import DeleteCommand
    from .manifest import ManifestCommand

    from .work_order import WorkOrderCommand
    from .model import ModelCommand



from ..utils import Command
from ...lib.site import SiteManager


class DeleteCommand(Command):
    """Delete a Site."""

    def setup(self, subparsers):
        parser = super(DeleteCommand, self).setup(subparsers)
        parser.add_argument('-i', '--site_id', required=True, type=str, help="Site id")
        return parser

    def run(self, site_id, **kwargs):
        return SiteManager().delete(site_id)



import io
import os
import cv2
import logging
try:
    import Queue as queue
except ImportError:
    import queue as queue

Full = queue.Full
Queue = queue.Queue
Empty = queue.Empty

LOGGER = logging.getLogger(__name__)
SUPPORTED_IMAGE_INPUT_FORMAT = ['.bmp', '.jpeg', '.jpg', '.jpe', '.png', '.tif', '.tiff']
SUPPORTED_VIDEO_INPUT_FORMAT = ['.avi', '.mp4', '.webm', '.mjpg']
SUPPORTED_FILE_INPUT_FORMAT = SUPPORTED_IMAGE_INPUT_FORMAT + SUPPORTED_VIDEO_INPUT_FORMAT
SUPPORTED_PROTOCOLS_INPUT = ['rtsp', 'http', 'https']
SUPPORTED_IMAGE_OUTPUT_FORMAT = SUPPORTED_IMAGE_INPUT_FORMAT
SUPPORTED_VIDEO_OUTPUT_FORMAT = ['.avi', '.mp4']

BGR_TO_COLOR_SPACE = {
    'RGB': cv2.COLOR_BGR2RGB,
    'YUV': cv2.COLOR_BGR2YUV,
    'YUV420': cv2.COLOR_BGR2YUV_I420,
    'HSV': cv2.COLOR_BGR2HSV,
    'GRAY': cv2.COLOR_BGR2GRAY,
}

SUPPORTED_VIDEO_OUTPUT_COLOR_SPACE = list(BGR_TO_COLOR_SPACE.keys()) + ['BGR']

SUPPORTED_FOURCC = {
    '.mp4': ['mp4v', 'avc1'],
    '.avi': ['XVID', 'MJPG']
}


class TqdmToLogger(io.StringIO):
    """Tqdm output stream to play nice with logger."""
    logger = None
    level = None
    buf = ''

    def __init__(self, logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip('\r\n\t ')

    def flush(self):
        self.logger.log(self.level, self.buf)


def clear_queue(queue):
    with queue.mutex:
        queue.queue.clear()


def write_frame_to_disk(frame, path):
    if frame.output_image is not None:
        if os.path.isfile(path):
            LOGGER.warning('File {} already exists. Skipping it.'.format(path))
        else:
            LOGGER.debug('Writing file {} to disk'.format(path))
            cv2.imwrite(path, frame.output_image)
    else:
        LOGGER.warning('No frame to output.')
    return



import errno
import requests
import json
import os
import os.path
import shutil
import subprocess

from git import Repo
from deepomatic.api.http_helper import HTTPHelper


DEEPOMATIC_SITE_PATH = os.path.join(os.path.expanduser('~'), '.deepomatic', 'sites')


def makedirs(folder, *args, **kwargs):
    # python2/3 compatible implementation of os.makedirs(exist_ok=True)
    # cf https://stackoverflow.com/a/60371380
    try:
        return os.makedirs(folder, exist_ok=True, *args, **kwargs)
    except TypeError:
        # Unexpected arguments encountered
        pass

    try:
        # Should work is TypeError was caused by exist_ok, eg., Py2
        return os.makedirs(folder, *args, **kwargs)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

        if os.path.isfile(folder):
            # folder is a file, raise OSError just like os.makedirs() in Py3
            raise


class SiteManager(object):
    def __init__(self, path=DEEPOMATIC_SITE_PATH, client_cls=HTTPHelper):
        makedirs(path)
        self._repo = Repo.init(path)
        self._client = client_cls()

        api_key = os.getenv('DEEPOMATIC_API_KEY')
        self.session = requests.Session()
        self.session.headers.update({'Content-type': 'application/json',
                                     'Authorization': 'Token {}'.format(api_key)})

    def get(self, site_id):
        return self._client.get('/sites/{}'.format(site_id))

    def create(self, name, description, app_version_id):
        data = {
            'name': name,
            'app_version_id': app_version_id
        }
        if description is not None:
            data['desc'] = description

        ret = self._client.post('/sites', data=data)
        return "New site created with id: {}".format(ret['id'])

    def update(self, site_id, app_version_id):
        data = {
            'app_version_id': app_version_id
        }

        ret = self._client.patch('/sites/{}'.format(site_id), data=data)
        return "Site {} updated".format(ret['id'])

    def delete(self, site_id):
        self._client.delete('/sites/{}'.format(site_id))
        return "Site {} deleted".format(site_id)

    def current(self):
        return str(self._repo.head.reference)

    def current_app(self):
        return self._repo.head.commit.message

    def get_deployment_manifest(self, site_id, target):
        uri = '/sites/{}/deployment-manifest?target={}'.format(site_id, target)
        return self._client.get(uri).decode()

    def install(self, site_id):
        site = self.get(site_id)
        docker_compose = self.get_deployment_manifest(site_id, 'docker-compose')
        try:
            # create branch
            self._repo.git.checkout(site_id, orphan=True)
        except Exception:
            return None

        docker_compose_path = os.path.join(self._repo.working_tree_dir, 'docker-compose.yml')
        with open(docker_compose_path, 'w') as f:
            f.write(docker_compose)

        # add to index
        self._repo.index.add([docker_compose_path])

        # commit
        self._repo.index.commit(site['app']['id'])

        return site['app']['id']

    def list(self):
        return [head.name for head in self._repo.heads]

    def load(self, archive_path):
        self._repo.git.bundle('verify', archive_path)
        site_id = self._repo.git.bundle('list-heads', archive_path).split('/')[-1]
        return self._repo.git.fetch(archive_path, '{}:{}'.format(site_id, site_id))

    def logs(self):
        return self._run_docker_compose("logs", "-f", "-t")

    def rollback(self, n=1):
        self._repo.git.reset("HEAD~{}".format(n), hard=True)
        return self._repo.head.commit.committed_datetime

    def save(self, site_id, archive_path):
        return self._repo.git.bundle('create', archive_path, site_id)

    def start(self):
        return self._run_docker_compose("up", "-d")

    def status(self):
        return self._run_docker_compose("ps")

    def stop(self):
        return self._run_docker_compose("down", "-v")

    def uninstall(self, site_id):
        try:
            self._repo.git.branch(site_id, D=True)
        except Exception:
            # last site, we can delete the repo
            shutil.rmtree(self._repo.working_tree_dir)

    def upgrade(self):
        site_id = self.current()
        site = self.get(site_id)

        if (self._repo.head.commit.message == site['app']['id']):
            return

        docker_compose = self.get_deployment_manifest(site_id, 'docker-compose')

        docker_compose_path = os.path.join(self._repo.working_tree_dir, 'docker-compose.yml')
        with open(docker_compose_path, 'w') as f:
            f.write(docker_compose)

        # add to index
        self._repo.index.add([docker_compose_path])

        # commit
        self._repo.index.commit(site['app']['id'])

        return site['app']['id']

    def use(self, site_id):
        if hasattr(self._repo.heads, site_id):
            getattr(self._repo.heads, site_id).checkout()
            return True
        else:
            # site does not exist
            return False

    def _run_docker_compose(self, *args):
        site_id = self.current()
        command = ' '.join(["docker-compose", "-f", os.path.join(DEEPOMATIC_SITE_PATH, "docker-compose.yml")] + list(args))
        env = {
            "DEEPOMATIC_SITE_ID": site_id
        }
        p = subprocess.Popen(command, env=env, shell=True)
        try:
            p.wait()
        except KeyboardInterrupt:
            try:
                p.terminate()
            except OSError:
                pass
            p.wait()

    def make_work_order_url(self, base_url):
        return "{}/v0.2/work-orders".format(base_url)

    def create_work_order(self, base_url, name, metadata):
        work_order_url = self.make_work_order_url(base_url)
        data = {
            'name': name,
            'metadata': metadata
        }
        res = self.session.post(work_order_url + '/', data=json.dumps(data))
        if res.status_code == 201:
            return res.json()['id']
        else:
            return res.text

    def status_work_order(self, base_url, work_order_id):
        work_order_url = self.make_work_order_url(base_url)
        res = self.session.get('{}/{}'.format(work_order_url, work_order_id))
        if res.status_code == 200:
            return res.json()
        else:
            return res.text

    def delete_work_order(self, base_url, work_order_id):
        work_order_url = self.make_work_order_url(base_url)
        res = self.session.delete('{}/{}'.format(work_order_url, work_order_id))
        if res.status_code == 204:
            return "Work order deleted"
        else:
            return res.text

    def make_work_order_inference(self, base_url, work_order_id, entries, metadata):
        work_order_url = self.make_work_order_url(base_url)
        input_data_url = "{}/{}/analyze".format(work_order_url, work_order_id)
        data = {
            'inputs': entries,
            "metadata": metadata
        }
        res = self.session.post(input_data_url + '/', data=json.dumps(data))
        if res.status_code == 200:
            return res.json()
        else:
            return res.text



from .utils import _CurrentSiteCommand, SiteManager


class UpgradeCommand(_CurrentSiteCommand):
    """
        Upgrade a site
    """

    def run(self, **kwargs):
        manager = SiteManager()
        site_id = manager.current()
        app_id = manager.upgrade()
        if app_id is None:
            print('Site is up-to-date')
        else:
            print('Site', site_id, 'upgraded with app', app_id)



from ...utils import Command


class ModelCommand(Command):
    """
        Model related commands.
    """

    from .noop import NoopCommand
    from .draw import DrawCommand
    from .blur import BlurCommand
    from .infer import InferCommand



import json
import logging
from .workflow_abstraction import AbstractWorkflow
from ..json_schema import validate_json, JSONSchemaType
from ..cmds.studio_helpers.vulcan2studio import transform_json_from_studio_to_vulcan
from ..exceptions import DeepoPredictionJsonError, DeepoOpenJsonError, SendInferenceError


LOGGER = logging.getLogger(__name__)


class JsonRecognition(AbstractWorkflow):

    class InferResult(AbstractWorkflow.AbstractInferResult):
        def __init__(self, frame_name, frame_pred):
            self.frame_name = frame_name
            self.frame_pred = frame_pred

        def get_predictions(self, timeout):
            return self.frame_pred

    def __init__(self, recognition_version_id, pred_file):
        super(JsonRecognition, self).__init__('r{}'.format(recognition_version_id))
        self._id = recognition_version_id
        self._pred_file = pred_file

        # Load the json
        try:
            with open(pred_file) as json_file:
                vulcan_json_with_pred = json.load(json_file)
        except Exception:
            raise DeepoOpenJsonError("Prediction JSON file {} is not a valid JSON file".format(pred_file))

        # Check json validity
        is_valid, error, schema_type = validate_json(vulcan_json_with_pred)
        if is_valid:
            if schema_type == JSONSchemaType.VULCAN:
                LOGGER.debug("Vulcan prediction JSON {} validated".format(pred_file))
            elif schema_type == JSONSchemaType.STUDIO:
                vulcan_json_with_pred = transform_json_from_studio_to_vulcan(vulcan_json_with_pred)
                LOGGER.debug("Studio prediction JSON {} validated and transformed to Vulcan format".format(pred_file))
        elif error is not None:
            LOGGER.warning("Error with {} JSON : {} in the instance {}".format(schema_type, error.message, list(error.path)))
            raise DeepoPredictionJsonError("Upload JSON file {} is not a proper {} JSON file".format(pred_file, schema_type))
        else:
            raise DeepoPredictionJsonError("Prediction JSON file {} is neither a proper Studio or Vulcan JSON file".format(pred_file))

        # Store predictions for easy access
        self._all_predictions = {vulcan_pred['data']['framename']: vulcan_pred for vulcan_pred in vulcan_json_with_pred}

    def close(self):
        pass

    def infer(self, _useless_encoded_image_bytes, _useless_push_client, frame_name):
        # _useless_encoded_image_bytes and _useless_push_client are used only for rpc and cloud workflows
        try:
            frame_pred = self._all_predictions[frame_name]
        except KeyError:
            raise SendInferenceError("Could not find predictions for frame {}".format(frame_name))

        return self.InferResult(frame_name, frame_pred)



from .thread_base import CurrentMessages


class Frame(object):
    def __init__(self, name, filename, image, decoded_video_frame_index=None, absolute_video_frame_index=None):
        # The Frame object is used as a data exchanged in the different queues
        self.name = name  # name of the frame
        self.filename = filename  # the original filename from which the frame was extracted
        self.image = image  # an opencv loaded image (numpy array)
        self.decoded_video_frame_index = decoded_video_frame_index  # frame index in the input sequence (skipped frames are not included)
        self.absolute_video_frame_index = absolute_video_frame_index  # frame index in the input sequence (skipped frames included)
        self.frame_number = None  # frame_number since deepocli started (set by input_loop)
        self.inference_async_result = None  # an inference request object that will allow us to retrieve the predictions when ready
        self.predictions = None  # predictions result dict
        self.output_image = None  # frame to output (modified version of the image, check infer postprocessings draw/blur)
        self.buf_bytes = None

    def __str__(self):
        return "<Frame {}>".format(' '.join("{}={}".format(key, getattr(self, key)) for key in [
            'name',
            'filename',
            'frame_number',
            'decoded_video_frame_index',
            'absolute_video_frame_index',
            'inference_async_result']))


class CurrentFrames(CurrentMessages):
    def forget_frame(self, frame, count_as_error=True):
        self.forget_message(frame.frame_number, count_as_error=count_as_error)

    def add_frame(self, frame):
        self.add_message(frame.frame_number)

    def get_oldest(self):
        return self.get_min()

    def pop_oldest(self):
        return self.pop_min()



from ..utils import Command
from ...lib.site import SiteManager


class CreateCommand(Command):
    """Create a new site."""

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-n', '--name', required=True, type=str, help="Site name")
        parser.add_argument('-d', '--description', type=str, help="Site description")
        parser.add_argument('-v', '--drive_app_version_id', required=True, type=str, help="DriveAppVersion id")
        return parser

    def run(self, name, description, drive_app_version_id, **kwargs):
        return SiteManager().create(name, description, drive_app_version_id)




def transform_json_from_vulcan_to_studio(vulcan_json):
    """Transforms a json from the vulcan format to the studio format."""
    # Initialize variables
    studio_json = {'tags': [], 'images': []}
    unique_tags = set()

    # Transform vulcan json to a list of vulcan json if needed
    if isinstance(vulcan_json, dict):
        vulcan_json = [vulcan_json]

    # Loop through all vulcan images
    for vulcan_image in vulcan_json:
        # Initialize studio images
        studio_image = {'annotated_regions': [], 'location': vulcan_image.get('location', '')}
        if 'data' in vulcan_image:
            studio_image['data'] = vulcan_image['data']

        # Loop through all vulcan predictions
        all_predictions = vulcan_image['outputs'][0]['labels']['predicted'] + vulcan_image['outputs'][0]['labels']['discarded']
        for prediction in all_predictions:
            # Build studio annotation in case of classification or tagging
            annotation = {
                "tags": [prediction['label_name']],
                "region_type": "Whole",
                "score": prediction['score'],
                "threshold": prediction['threshold']
            }

            # Add bounding box if needed
            if 'roi' in prediction:
                annotation['region_type'] = 'Box'
                annotation['region'] = {
                    "xmin": prediction['roi']['bbox']['xmin'],
                    "xmax": prediction['roi']['bbox']['xmax'],
                    "ymin": prediction['roi']['bbox']['ymin'],
                    "ymax": prediction['roi']['bbox']['ymax']
                }

            # Update json and unique tags
            studio_image['annotated_regions'].append(annotation)
            unique_tags.add(prediction['label_name'])

        # Update studio json
        studio_json['images'].append(studio_image)

    # Update final unique tags
    studio_json['tags'] = list(unique_tags)

    return studio_json


def transform_json_from_studio_to_vulcan(studio_json):
    """Transforms a json from the studio format to the vulcan format."""
    # Initialize variables
    vulcan_json = []

    # Loop through all studio images
    for studio_image in studio_json['images']:
        # Initialize vulcan prediction
        vulcan_pred = {'outputs': [{'labels': {'discarded': [], 'predicted': []}}]}
        predicted = []
        discarded = []
        for metadata in ['location', 'data']:
            if metadata in studio_image:
                vulcan_pred[metadata] = studio_image[metadata]

        # Loop through all studio predictions
        for studio_pred in studio_image['annotated_regions']:
            # Build vulcan annotation
            annotation = {
                'label_name': studio_pred['tags'][0],
                'score': studio_pred['score'],
                'threshold': studio_pred['threshold']
            }

            # Add bounding box if needed
            if studio_pred['region_type'] == 'Box':
                annotation['roi'] = {
                    'bbox': {
                        'xmin': studio_pred['region']['xmin'],
                        'xmax': studio_pred['region']['xmax'],
                        'ymin': studio_pred['region']['ymin'],
                        'ymax': studio_pred['region']['ymax']
                    }
                }

            # Update json
            if annotation['score'] >= annotation['threshold']:
                predicted.append(annotation)
            else:
                discarded.append(annotation)

        # Sort by prediction score of descending order
        predicted = sorted(predicted, key=lambda k: k['score'], reverse=True)
        discarded = sorted(discarded, key=lambda k: k['score'], reverse=True)
        vulcan_pred['outputs'][0]['labels']['predicted'] = predicted
        vulcan_pred['outputs'][0]['labels']['discarded'] = discarded

        # Update vulcan json
        vulcan_json.append(vulcan_pred)

    return vulcan_json



from deepomatic.cli.cmds.utils import Command
from ..utils import InferManager, BlurImagePostprocessing
from deepomatic.cli.cmds.utils import setup_model_cmd_line_parser


class BlurCommand(Command):
    """
        Generates new images and videos with predictions results blurred on them. Computes prediction if JSON has not yet been generated.
        Typical usage is: deepo platform model blur -i img.png -o pred.json draw.png -r 12345
    """

    def setup(self, subparsers):
        parser = super(BlurCommand, self).setup(subparsers)

        setup_model_cmd_line_parser("platform", "blur", parser)
        return parser

    def run(self, **kwargs):
        return InferManager().input_loop(kwargs, BlurImagePostprocessing(**kwargs))



from ...utils import PlatformCommand
from ..utils import DrivePlatformManager


class UpdateCommand(PlatformCommand):
    """Update an existing DriveAppVersion."""

    def setup(self, subparsers):
        parser = super(UpdateCommand, self).setup(subparsers)
        parser.add_argument('-v', '--drive_app_version_id', required=True, type=str, help="DriveAppVersion id")
        parser.add_argument('-n', '--name', type=str, help="DriveAppVersion name")
        parser.add_argument('-d', '--description', type=str, help="DriveAppVersion description")
        return parser

    def run(self, drive_app_version_id, name, description, **kwargs):
        return DrivePlatformManager().update_app_version(drive_app_version_id, name, description)



import os
import sys
import logging
import argparse
import argcomplete
from .cmds import AVAILABLE_COMMANDS
from .cmds.utils import CommandResult
from .version import __version__, __title__


class ParserWithHelpOnError(argparse.ArgumentParser):
    """
    Modifies argparser to display the help whenever an error is triggered.
    """

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(1)


def argparser_init():
    # Initialize main argparser and version command
    argparser = ParserWithHelpOnError(prog='deepo')
    argparser.add_argument(
        '-v', '--version', action='version',
        version='{title} {version}'.format(title=__title__, version=__version__)
    )

    subparsers = argparser.add_subparsers(dest='command', help='')
    subparsers.required = True
    return argparser, subparsers


def run(args):
    # deprecated commands (TODO: migrate to RunCommand() ?)
    argparser, subparsers = argparser_init()

    for command in AVAILABLE_COMMANDS:
        command.setup(subparsers)

    subparsers.required = True
    argcomplete.autocomplete(argparser)

    args = argparser.parse_args(args)

    # Update the log level accordingly
    if args.verbose:
        log_level = logging.DEBUG
        log_format = '[%(levelname)s %(name)s %(asctime)s %(process)d %(thread)d %(filename)s:%(lineno)s] %(message)s'
    else:
        log_level = os.getenv('DEEPOMATIC_LOG_LEVEL', logging.INFO)
        log_format = '[%(levelname)s %(asctime)s] %(message)s'
    logging.basicConfig(level=log_level, format=log_format)

    result = args.func(vars(args))

    if result:
        if isinstance(result, CommandResult) and args.json_output:
            print(result.to_json_str(indent=args.json_output_indent))
        else:
            print(result)

    return result



from ..utils import Command
from deepomatic.cli.lib.system import SystemManager


class SystemCommand(Command):
    """
        System command
    """

    class CheckCommand(Command):
        """
            Check the system
        """

        def run(self, **kwargs):
            SystemManager().check()



from deepomatic.cli.cmds.utils import Command
from ..utils import InferManager
from deepomatic.cli.cmds.utils import setup_model_cmd_line_parser


class InferCommand(Command):
    """
        Computes prediction on a file or directory and outputs results as a JSON file.
        Typical usage is: deepo platform model infer -i img.png -o pred.json -r 12345
    """

    def setup(self, subparsers):
        parser = super(InferCommand, self).setup(subparsers)
        setup_model_cmd_line_parser("platform", "infer", parser)
        return parser

    def run(self, **kwargs):
        return InferManager().input_loop(kwargs)



import os


class AbstractWorkflow(object):
    class AbstractInferResult(object):
        def get_predictions(self):
            raise NotImplementedError()

        def __str__(self):
            return '<{}>'.format(self.__class__.__qualname__)

    def __init__(self, display_id):
        self._display_id = display_id

    def new_client(self):
        return None

    def close_client(self, client):
        pass

    def close(self):
        raise NotImplementedError()

    @property
    def display_id(self):
        return self._display_id

    def infer(self, encoded_image_bytes, push_client):
        """Should return a subclass of AbstractInferResult"""
        raise NotImplementedError()

    def get_json_output_filename(self, file):
        dirname = os.path.dirname(file)
        filename, _ = os.path.splitext(file)
        return os.path.join(dirname, filename + '.{}.json'.format(self.display_id))



import os
import cv2
import logging

from .common import (SUPPORTED_IMAGE_INPUT_FORMAT, SUPPORTED_PROTOCOLS_INPUT,
                     SUPPORTED_VIDEO_INPUT_FORMAT, clear_queue)
from .exceptions import DeepoFPSError, DeepoInputError, DeepoVideoOpenError
from .frame import Frame
from .thread_base import Thread


LOGGER = logging.getLogger(__name__)


def get_input(descriptor, kwargs):
    if descriptor is None:
        raise DeepoInputError('No input specified. use -i flag')
    elif os.path.exists(descriptor):
        if os.path.isfile(descriptor):
            # Single image file
            if ImageInputData.is_valid(descriptor):
                LOGGER.debug('Image input data detected for {}'.format(descriptor))
                return ImageInputData(descriptor, **kwargs)
            # Single video file
            elif VideoInputData.is_valid(descriptor):
                LOGGER.debug('Video input data detected for {}'.format(descriptor))
                return VideoInputData(descriptor, **kwargs)
            else:
                raise DeepoInputError('Unsupported input file type')
        # Input directory containing images, videos, or json
        elif os.path.isdir(descriptor):
            LOGGER.debug('Directory input data detected for {}'.format(descriptor))
            return DirectoryInputData(descriptor, **kwargs)
        else:
            raise DeepoInputError('Unknown input path')
    # Device indicated by digit number such as a webcam
    elif descriptor.isdigit():
        LOGGER.debug('Device input data detected for {}'.format(descriptor))
        return DeviceInputData(descriptor, **kwargs)
    # Video stream such as RTSP
    elif StreamInputData.is_valid(descriptor):
        LOGGER.debug('Stream input data detected for {}'.format(descriptor))
        return StreamInputData(descriptor, **kwargs)
    else:
        raise DeepoInputError('Unknown input')


class InputThread(Thread):
    def __init__(self, exit_event, input_queue, output_queue, inputs):
        super(InputThread, self).__init__(exit_event, input_queue, output_queue)
        self.inputs = inputs
        self.frame_number = 0  # Used to keep input order, notably for video reconstruction

    def process_msg(self, _unused):
        try:
            frame = next(self.inputs)
        except StopIteration:
            self.stop()
            return

        if self.inputs.is_infinite():
            # Discard all previous inputs
            clear_queue(self.output_queue)

        frame.frame_number = self.frame_number
        # TODO: for a stream put should not be blocking
        return frame

    def put_to_output(self, msg):
        super(InputThread, self).put_to_output(msg)
        self.frame_number += 1


class InputData(object):
    def __init__(self, descriptor, **kwargs):
        self._args = kwargs
        self._descriptor = descriptor
        self._filename = str(descriptor)
        self._name = os.path.basename(os.path.normpath(self._filename))
        base, ext = os.path.splitext(self._name)
        if ext:
            self._name = '{}_{}'.format(base, ext.lstrip('.'))
        recognition_id = kwargs.get('recognition_id', '')
        self._reco = '' if recognition_id is None else recognition_id

    def __iter__(self):
        raise NotImplementedError()

    def __next__(self):
        raise NotImplementedError()

    def next(self):
        return self.__next__()  # for python 2

    def get_fps(self):
        raise NotImplementedError()

    def get_frame_count(self):
        raise NotImplementedError()

    def is_infinite(self):
        raise NotImplementedError()


class ImageInputData(InputData):
    @classmethod
    def is_valid(cls, descriptor):
        _, ext = os.path.splitext(descriptor)
        return os.path.exists(descriptor) and ext.lower() in SUPPORTED_IMAGE_INPUT_FORMAT

    def __init__(self, descriptor, **kwargs):
        super(ImageInputData, self).__init__(descriptor, **kwargs)
        self._name = '%s_%s' % (self._name, self._reco)

    def __iter__(self):
        self._iterator = iter([Frame(self._name, self._filename, cv2.imread(self._descriptor, 1))])
        return self

    def __next__(self):
        return next(self._iterator)

    def get_fps(self):
        return 0

    def get_frame_count(self):
        return 1

    def is_infinite(self):
        return False


class VideoInputData(InputData):
    @classmethod
    def is_valid(cls, descriptor):
        _, ext = os.path.splitext(descriptor)
        return os.path.exists(descriptor) and ext.lower() in SUPPORTED_VIDEO_INPUT_FORMAT

    def __init__(self, descriptor, **kwargs):
        super(VideoInputData, self).__init__(descriptor, **kwargs)
        self._absolute_video_frame_index = 0
        self._decoded_video_frame_index = 0
        self._name = '%s_%s_%s' % (self._name, '%05d', self._reco)
        self._cap = None
        self._open_video()
        self._kwargs_fps = kwargs['input_fps']
        self._skip_frame = kwargs['skip_frame']
        self._extract_fps = None
        self._fps = self.get_fps()

    def _open_video(self, raise_exc=True):
        if self._cap is not None:
            self._cap.release()
        self._cap = cv2.VideoCapture(self._descriptor)

        if not self._cap.isOpened():
            self._cap = None
            if raise_exc:
                raise DeepoVideoOpenError("Could not open video {}".format(self._descriptor))
            return False
        return True

    def __iter__(self):
        self._open_video()
        self._absolute_video_frame_index = 0
        self._decoded_video_frame_index = 0
        self._frames_to_skip = 0
        self._should_skip_fps = self._video_fps
        return self

    def _stop_video(self, raise_exc=True):
        self._cap.release()
        if raise_exc:
            raise StopIteration()

    def _grab_next(self):
        grabbed = self._cap.grab()
        if not grabbed:
            self._stop_video()
        else:
            self._absolute_video_frame_index += 1

    def _decode_next(self):
        decoded, frame = self._cap.retrieve()
        if not decoded:
            self._stop_video()
        else:
            self._decoded_video_frame_index += 1
            return Frame(self._name % self._absolute_video_frame_index,
                         self._filename, frame,
                         self._decoded_video_frame_index,
                         self._absolute_video_frame_index)

    def _read_next(self):
        read, frame = self._cap.read()
        if read:
            self._absolute_video_frame_index += 1
            self._decoded_video_frame_index += 1
            return Frame(self._name % self._absolute_video_frame_index,
                         self._filename, frame,
                         self._decoded_video_frame_index,
                         self._absolute_video_frame_index)
        else:
            self._stop_video()

    def __next__(self):
        # make sure we don't enter infinite loop
        assert self._frames_to_skip >= 0
        assert self._extract_fps >= 0

        while True:
            # first, check if the frame should be skipped because of extract fps
            if self._extract_fps > 0:
                if self._should_skip_fps < self._video_fps:
                    self._grab_next()
                    self._should_skip_fps += self._extract_fps
                    continue
                else:
                    self._should_skip_fps += self._extract_fps - self._video_fps

            # then, check if the frame should be skipped because of skipped frame
            if self._frames_to_skip:
                self._grab_next()
                self._frames_to_skip -= 1
                continue
            else:
                self._frames_to_skip = self._skip_frame

            return self._read_next()

    def get_fps(self):
        # There are three different type of fps:
        #   _video_fps: original video fps
        #   _kwarg_fps: fps specified by the user through the CLI if any
        #   _extract_fps: fps used for frame extraction
        assert(self._cap is not None)
        # Retrieve the original video fps if available
        try:
            self._video_fps = self._cap.get(cv2.CAP_PROP_FPS)
        except Exception:
            raise DeepoFPSError('Could not read fps for video {}, please specify it with --input_fps option.'.format(self._descriptor))
        if self._video_fps == 0:
            raise DeepoFPSError('Null fps detected for video {}, please specify it with --input_fps option.'.format(self._descriptor))

        # Compute fps for frame extraction so that we don't analyze useless frame that will be discarded later
        if not self._kwargs_fps:
            self._extract_fps = self._video_fps
            LOGGER.debug('No --input_fps specified, using raw video fps of {}'.format(self._video_fps))
        elif self._kwargs_fps < self._video_fps:
            self._extract_fps = self._kwargs_fps
            LOGGER.debug('Using user-specified --input_fps of {} instead of raw video fps of {}'.format(self._kwargs_fps, self._video_fps))
        else:
            self._extract_fps = self._video_fps
            LOGGER.debug('User-specified --input_fps of {} specified'
                         ' but using maximum raw video fps of {}'.format(self._kwargs_fps, self._video_fps))

        return self._extract_fps

    def get_frame_count(self):
        assert self._video_fps > 0

        fps_ratio = self._extract_fps / self._video_fps
        skip_ratio = 1. / (1 + self._skip_frame)
        try:
            return int(self._cap.get(cv2.CAP_PROP_FRAME_COUNT) * fps_ratio * skip_ratio)
        except Exception:
            LOGGER.warning('Cannot compute the total frame count')
            return 0

    def is_infinite(self):
        return False


class DirectoryInputData(InputData):
    @classmethod
    def is_valid(cls, descriptor):
        return (os.path.exists(descriptor) and os.path.isdir(descriptor))

    def __init__(self, descriptor, **kwargs):
        super(DirectoryInputData, self).__init__(descriptor, **kwargs)
        self._current = None
        self._files = []
        self._inputs = []
        self._recursive = self._args['recursive']

        if self.is_valid(descriptor):
            _paths = [os.path.join(descriptor, name) for name in os.listdir(descriptor)]
            self._inputs = []
            for path in sorted(_paths):
                if ImageInputData.is_valid(path):
                    LOGGER.debug('Image input data detected for {}'.format(path))
                    self._inputs.append(ImageInputData(path, **kwargs))
                elif VideoInputData.is_valid(path):
                    LOGGER.debug('Video input data detected for {}'.format(path))
                    self._inputs.append(VideoInputData(path, **kwargs))
                elif self._recursive and self.is_valid(path):
                    LOGGER.debug('Directory input data detected for {}'.format(path))
                    self._inputs.append(DirectoryInputData(path, **kwargs))

    def _gen(self):
        for source in self._inputs:
            for frame in source:
                yield frame

    def __iter__(self):
        self.gen = self._gen()
        return self

    def __next__(self):
        return next(self.gen)

    def get_frame_count(self):
        return sum([_input.get_frame_count() for _input in self._inputs])

    def get_fps(self):
        return 1

    def is_infinite(self):
        return False


class StreamInputData(VideoInputData):
    @classmethod
    def is_valid(cls, descriptor):
        return '://' in descriptor and descriptor.split('://')[0].lower() in SUPPORTED_PROTOCOLS_INPUT

    def __init__(self, descriptor, **kwargs):
        super(StreamInputData, self).__init__(descriptor, **kwargs)
        self._name = 'stream_%s_%s' % ('%05d', self._reco)

    def get_frame_count(self):
        return -1

    def is_infinite(self):
        return True


class DeviceInputData(VideoInputData):

    @classmethod
    def is_valid(cls, descriptor):
        return descriptor.isdigit()

    def __init__(self, descriptor, **kwargs):
        super(DeviceInputData, self).__init__(int(descriptor), **kwargs)
        self._name = 'device%s_%s_%s' % (descriptor, '%05d', self._reco)

    def get_frame_count(self):
        return -1

    def is_infinite(self):
        return True



from .utils import _CurrentSiteCommand, SiteManager


class RollbackCommand(_CurrentSiteCommand):
    """
        Rollback a site
    """

    def setup(self, subparsers):
        parser = super(RollbackCommand, self).setup(subparsers)
        parser.add_argument('n', nargs='?', type=int, help="how far back to rollback", default=1)
        return parser

    def run(self, n, **kwargs):
        print(SiteManager().rollback(n))



from deepomatic.cli.lib.site import SiteManager  # noqa: F401
from ..utils import Command


class _SiteCommand(Command):
    """
        Commands that expect a site_id
    """

    def setup(self, subparsers):
        parser = super(_SiteCommand, self).setup(subparsers)
        parser.add_argument('site_id', type=str, help="site id")
        return parser


class _CurrentSiteCommand(Command):
    """
        Commands that should work on the current site
    """



from ...utils import PlatformCommand
from ..utils import EngagePlatformManager


class DeleteCommand(PlatformCommand):
    """Delete an EngageApp."""

    def setup(self, subparsers):
        parser = super(DeleteCommand, self).setup(subparsers)
        parser.add_argument('-i', '--engage_app_id', required=True, type=str, help="EngageApp id")
        return parser

    def run(self, engage_app_id, **kwargs):
        return EngagePlatformManager().delete_app(engage_app_id)



from .utils import _SiteCommand, SiteManager


class SaveCommand(_SiteCommand):
    """
        Save a site to an archive
    """

    def setup(self, subparsers):
        parser = super(SaveCommand, self).setup(subparsers)
        parser.add_argument('archive_dst', type=str, help="path to archive")
        return parser

    def run(self, site_id, archive_dst, **kwargs):
        SiteManager().save(site_id, archive_dst)
        # TODO: feedback



from deepomatic.cli.lib.platform import DrivePlatformManager  # noqa: F401
from deepomatic.cli.lib.platform import EngagePlatformManager  # noqa: F401
from deepomatic.cli.lib.inference import InferManager, DrawImagePostprocessing, BlurImagePostprocessing  # noqa: F401
from deepomatic.cli.lib.add_images import AddImageManager  # noqa: F401



from ..utils import Command


class PlatformCommand(Command):
    """
        Operations on the Deepomatic Platform (studio)
    """

    from .drive_app import DriveAppCommand
    from .drive_app_version import DriveAppVersionCommand
    from .engage_app import EngageAppCommand
    from .engage_app_version import EngageAppVersionCommand
    from .model import ModelCommand
    from .add_images import AddImagesCommand



from .workflow_abstraction import AbstractWorkflow
from ..exceptions import ResultInferenceError, ResultInferenceTimeout
from ..exceptions import DeepoRPCRecognitionError, DeepoRPCUnavailableError


def import_rpc_package(should_raise=False):
    rpc, protobuf = None, None
    try:
        from deepomatic import rpc
    except ImportError:
        if should_raise:
            raise DeepoRPCUnavailableError('RPC not installed')

    try:
        from deepomatic.rpc import v07_ImageInput, BINARY_IMAGE_PREFIX  # noqa: F401
        from deepomatic.rpc.client import Client  # noqa: F401
        from deepomatic.rpc.exceptions import ServerError  # noqa: F401
        from deepomatic.rpc.amqp.exceptions import Timeout  # noqa: F401
        from deepomatic.rpc.response import wait_responses  # noqa: F401
        from deepomatic.rpc.helpers.v07_proto import create_recognition_command_mix  # noqa: F401
        from deepomatic.rpc.helpers.proto import create_v07_images_command  # noqa: F401
        from deepomatic.rpc.buffers.protobuf.cli.Message_pb2 import Message  # noqa: F401
        from google import protobuf  # noqa: F401
        import google.protobuf.json_format  # noqa: F401
    except ImportError as e:
        if should_raise:
            raise DeepoRPCUnavailableError('RPC not up-to-date: %s' % e)

    return rpc, protobuf


rpc, protobuf = import_rpc_package()


# decorator that prevents class instanciation when deepomatic rpc is not available
def requires_deepomatic_rpc(cls):
    def override_init(old_init):
        def __init__(self, *args, **kwargs):
            # checks that rpc is installed and up-to-date
            import_rpc_package(should_raise=True)
            try:
                old_init(self, *args, **kwargs)
            except Exception:
                old_init(self)

        return __init__
    cls.__init__ = override_init(cls.__init__)
    return cls


@requires_deepomatic_rpc
class RpcRecognition(AbstractWorkflow):
    @requires_deepomatic_rpc
    class InferResult(AbstractWorkflow.AbstractInferResult):
        def __init__(self, correlation_id, consumer):
            self._correlation_id = correlation_id
            self._consumer = consumer

        def get_predictions(self, timeout):
            try:
                response = self._consumer.get(self._correlation_id, timeout=timeout)
                try:
                    outputs = response.to_parsed_result_buffer()
                    predictions = {
                        'outputs': [
                            {
                                'labels': protobuf.json_format.MessageToDict(output.labels,
                                                                             including_default_value_fields=True,
                                                                             preserving_proto_field_name=True)
                            } for output in outputs
                        ]
                    }
                    return predictions
                except rpc.exceptions.ServerError as e:
                    raise ResultInferenceError({'error': str(e), 'code': e.code})
            except rpc.amqp.exceptions.Timeout:
                raise ResultInferenceTimeout(timeout)

        def __str__(self):
            return '<{} correlation_id={}>'.format(self.__class__.__qualname__, self._correlation_id)

    def __init__(self, recognition_version_id, amqp_url, routing_key, recognition_cmd_kwargs=None):
        super(RpcRecognition, self).__init__('recognition_{}'.format(recognition_version_id))
        self._id = recognition_version_id

        self._routing_key = routing_key
        self._consumer = None
        self.amqp_url = amqp_url

        # We declare the client that will be used for consuming in one thread only
        # RPC client is not thread safe
        self._consume_client = rpc.client.Client(amqp_url)
        if recognition_version_id is None:
            self._command_mix = rpc.helpers.v07_proto.create_workflow_command_mix()
        else:
            recognition_cmd_kwargs = recognition_cmd_kwargs or {'show_discarded': True, 'max_predictions': 1000}

            try:
                recognition_version_id = int(recognition_version_id)
            except ValueError:
                raise DeepoRPCRecognitionError("Cannot cast recognition ID into a number")

            self._command_mix = rpc.helpers.v07_proto.create_recognition_command_mix(recognition_version_id,
                                                                                     **recognition_cmd_kwargs)
        self._command_queue = self._consume_client.new_queue(self._routing_key)
        self._response_queue, self._consumer = self._consume_client.new_consuming_queue()

    def close_client(self, client):
        client.amqp_client.ensured_connection.close()

    def new_client(self):
        # Allow to create multiple clients for threads that will push
        # Since RPC client is not thread safe
        return rpc.client.Client(self.amqp_url)

    def close(self):
        if self._consumer is not None:
            self._consume_client.remove_consuming_queue(self._response_queue, self._consumer)
        self.close_client(self._consume_client)

    def infer(self, encoded_image_bytes, push_client, _useless_frame_name):
        # _useless_frame_name is used for the json workflow
        image_input = rpc.v07_ImageInput(source=rpc.BINARY_IMAGE_PREFIX + encoded_image_bytes)
        # forward_to parameter can be removed for images of worker nn with tag >= 0.7.8
        reply_to = self._response_queue.name
        serialized_buffer = rpc.helpers.proto.create_v07_images_command([image_input], self._command_mix, forward_to=[reply_to])
        correlation_id = push_client.send_binary(serialized_buffer, self._command_queue.name, reply_to=reply_to)
        return self.InferResult(correlation_id, self._consumer)



# -*- coding: utf-8 -*-
import gevent


class Task(object):
    def __init__(self, helper, pk=None, files=None):
        self._helper = helper

    def retrieve(self, task_id, wait=True):
        ret = self._helper.get('manage/tasks/{}/'.format(task_id))
        if not wait:
            return ret
        sleep_time = 0.3
        while ret['next'] is not None and ret['status'] != 'SUCCESS':
            if ret['status'] in ('FAILURE', 'REVOKED'):
                # TODO: use current_message.report_error()
                raise RuntimeError("Task {} stopped with status".format(task_id))
            elif ret['status'] == 'SUCCESS':
                task_id = ret['next']
                ret = self._helper.get('manage/tasks/{}/'.format(task_id))
            else:
                gevent.sleep(sleep_time)
                sleep_time = min(sleep_time + 0.3, 5)
                ret = self._helper.get('manage/tasks/{}/'.format(task_id))
        return ret



from ...utils import Command, BuildDict, valid_path
from ..utils import SiteManager

import argparse
import base64


def build_data(type, value):
    if type == "text":
        return value
    elif type == "image":
        if value.startswith(('http://', 'https://')):
            return value
        else:
            try:
                path = valid_path(value)
                with open(path, 'r') as f:
                    encoded_string = base64.b64encode(f.read())
                    return 'data:image/jpg;base64,' + encoded_string.decode('utf-8')
            except Exception:
                raise Exception('image {} not supported'.format(value))

    elif type == "number":
        return value
    else:
        raise Exception("type {} not supported".format(type))


class BuildCustomerAPIInput(argparse.Action):
    """
    This class is used in argparse. It will transform a chain of name@type@values into a list of inputs for
    the CustomerAPI.
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(BuildCustomerAPIInput, self).__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        entries = []
        for kv in values:
            vals = kv.split("@")
            assert len(vals) == 3

            name = vals[0]
            e_type = vals[1]
            value = vals[2]

            item = {
                'name': name,
                'data':
                {
                    'type': e_type,
                    'value': build_data(e_type, value),
                }
            }
            entries.append(item)
        setattr(namespace, self.dest, entries)


class InferCommand(Command):
    """
        Make an inference on a work order
    """

    def setup(self, subparsers):
        parser = super(InferCommand, self).setup(subparsers)
        parser.add_argument('-u', "--api_url", required=True, type=str, help="url of your Customer api")
        parser.add_argument('-i', '--work_order_id', required=True, type=str, help="Work order id")
        parser.add_argument('-e', "--entries", dest='entries',
                            action=BuildCustomerAPIInput, nargs="+", metavar="NAME@TYPE@VAL",
                            help='Inputs that will be passed to the CustomerAPI. They should be in this format name@type@value',
                            required=True)
        parser.add_argument('-m', "--metadata", dest='metadata', required=True,
                            action=BuildDict, nargs="+", metavar="NAME:VAL",
                            help='Metadata for the inference in format name:value')
        return parser

    def run(self, api_url, work_order_id, entries, metadata, **kwargs):
        return SiteManager().make_work_order_inference(api_url, work_order_id, entries, metadata)



from ..utils import Command
from ...lib.site import SiteManager


class UpdateCommand(Command):
    """
        Update an existing site
    """

    def setup(self, subparsers):
        parser = super(UpdateCommand, self).setup(subparsers)
        parser.add_argument('-i', '--site_id', required=True, type=str, help="Site id")
        parser.add_argument('-v', '--drive_app_version_id', required=True, type=str, help="DriveAppVersion id")
        return parser

    def run(self, site_id, drive_app_version_id, **kwargs):
        return SiteManager().update(site_id, drive_app_version_id)



from ....utils import Command
from deepomatic.cli.lib.camera import get_camera_ctrl


class _CameraCommand(Command):
    """
        Base class for commands that requires the camera's name
    """

    def setup(self, subparsers):
        parser = super(_CameraCommand, self).setup(subparsers)
        parser.add_argument('name', type=str, help="camera name")
        return parser


class CameraCommand(Command):
    """
        Control the Camera server
    """

    class AddCommand(Command):
        """
            Add a camera
        """

        def setup(self, subparsers):
            parser = super(CameraCommand.AddCommand, self).setup(subparsers)
            parser.add_argument('name', type=str, help="camera name")
            parser.add_argument('address', type=str, help="camera address")
            parser.add_argument('--fps', type=float, help="camera fps", default=None)

            group = parser.add_mutually_exclusive_group()
            group.add_argument("--tcp", action="store_true")
            group.add_argument("--udp", action="store_true")

            return parser

        def run(self, name, address, fps, tcp, udp, **kwargs):
            print(get_camera_ctrl().add(name, address))  # TODO: tcp/udp, fps

    class DeleteCommand(_CameraCommand):
        """
            Remove a camera
        """

        def run(self, name, **kwargs):
            print(get_camera_ctrl().delete(name))

    class StartCommand(_CameraCommand):
        """
            Start a camera
        """

        def run(self, name, **kwargs):
            print(get_camera_ctrl().start(name))

    class StopCommand(_CameraCommand):
        """
            Stop a camera
        """

        def run(self, name, **kwargs):
            print(get_camera_ctrl().stop(name))

    class ListCommand(Command):
        """
            List the cameras
        """

        def run(self, **kwargs):
            print(get_camera_ctrl().list())



from ...utils import PlatformCommand


class DriveAppVersionCommand(PlatformCommand):
    """DriveApp version related commands."""

    from .create import CreateCommand
    from .update import UpdateCommand
    from .delete import DeleteCommand



import os
import io
from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))


about = {}
with io.open(os.path.join(here, 'deepomatic', 'cli', 'version.py'), 'r', encoding='utf-8') as f:
    exec(f.read(), about)

with io.open(os.path.join(here, 'README.md'), 'r', encoding='utf-8') as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Read requirements
with io.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.readlines()

namespaces = ['deepomatic']

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    packages=find_packages(),
    package_dir={about['__title__']: 'deepomatic'},
    namespace_packages=namespaces,
    include_package_data=True,
    long_description=README,
    long_description_content_type='text/markdown',
    data_files=[('', ['requirements.txt'])],
    install_requires=requirements,
    extras_require={'rpc': ['deepomatic-rpc>=0.8.0']},
    python_requires=">=3.6.*",
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    entry_points={'console_scripts': ['deepo = deepomatic.cli.__main__:main']}
)



from ...utils import (
    PlatformCommand,
    valid_path,
)
from ..utils import EngagePlatformManager


class CreateCommand(PlatformCommand):
    """Create a new EngageAppVersion."""

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-w', '--workflow', required=True, type=valid_path, help="Path to the workflow yaml file")
        parser.add_argument('-c', '--custom_nodes', type=valid_path, help="Path to the custom nodes python file")
        parser.add_argument('-i', '--engage_app_id', required=True, type=str, help="EngageApp id for this EngageAppVersion")
        parser.add_argument('-p', '--base_major_version', default=None, type=str, help="Previous major EngageAppVersion version")
        parser.add_argument(
            '-r',
            '--recognition-version-ids',
            required=True,
            nargs="*",
            type=int,
            help="List of Recognition Version Id, one for each Recognition Spec in the EngageAppVersion version", default=[]
        )
        return parser

    def run(self,
            engage_app_id,
            workflow,
            custom_nodes,
            recognition_version_ids,
            base_major_version,
            **kwargs):

        return EngagePlatformManager().create_app_version(
            engage_app_id,
            workflow,
            custom_nodes,
            recognition_version_ids,
            base_major_version
        )



from ...utils import PlatformCommand
from ..utils import DrivePlatformManager


class UpdateCommand(PlatformCommand):
    """Update an existing DriveApp."""

    def setup(self, subparsers):
        parser = super(UpdateCommand, self).setup(subparsers)
        parser.add_argument('-i', '--drive_app_id', required=True, type=str, help="DriveApp id")
        parser.add_argument('-n', '--name', type=str, help="DriveApp name")
        parser.add_argument('-d', '--description', type=str, help="DriveApp description")
        return parser

    def run(self, drive_app_id, name, description, **kwargs):
        return DrivePlatformManager().update_app(drive_app_id, name, description)



from ..utils import Command
from .utils import SiteManager


class LoadCommand(Command):
    """
        Load a site from an archive
    """

    def setup(self, subparsers):
        parser = super(LoadCommand, self).setup(subparsers)
        parser.add_argument('archive', type=str, help="path to archive")
        return parser

    def run(self, archive, **kwargs):
        SiteManager().load(archive)



import os
import subprocess


class SystemManager(object):
    def check(self):
        print("=== Check git, deepomatic-cli installations")
        self.check_command(["git", "--version"], command_message="git")
        self.check_command(["deepo", "--version"], command_message="deepomatic-cli")

        print("=== Check docker, docker-compose and nvidia installation")
        self.check_command("docker", command_message="docker")
        self.check_command("nvidia-smi")
        self.check_command(["docker-compose", "--version"], command_message="docker-compose")
        self.check_command(["docker", "ps"], command_message="docker", success_message="is usable", error_message="is not usable")
        self.check_command(["docker", "run", "--gpus", "all", "nvidia/cuda:9.0-base", "nvidia-smi"], command_message="nvidia-docker")

        print("=== Check deepomatic api accessibility")
        self.check_command(["curl", "-Lf", "https://api.deepomatic.com"], command_message="Deepomatic API",
                           success_message="is reachable", error_message="is not reachable")

        print("=== Check deepomatic run docker images")
        self.check_command(["docker", "image", "inspect", "deepomatic/run-neural-worker:0.5.0-native"],
                           command_message="Neural worker image", success_message="is available", error_message="is not available")
        self.check_command(["docker", "image", "inspect", "deepomatic/run-workflow-server:0.5.0"],
                           command_message="Workflow server image", success_message="is available", error_message="is not available")
        self.check_command(["docker", "image", "inspect", "deepomatic/run-camera-server:0.5.0"],
                           command_message="Camera server image", success_message="is available", error_message="is not available")

    def check_command(self, command,
                      success_icon="\u2713", error_icon="\u2717",
                      command_message=None, success_message="is installed", error_message="is not installed"):
        try:
            if command_message is None:
                command_message = command if isinstance(command, str) else ' '.join(command)
            with open(os.devnull, "w") as null:
                result = subprocess.call(command, stdout=null, stderr=null)
            if result != 0:
                raise OSError
            print(success_icon, '\t', '\033[1m' + command_message + '\033[0m', success_message)
        except OSError:
            print(error_icon, '\t', '\033[1m' + command_message + '\033[0m', error_message)



from ...utils import Command
from ..utils import SiteManager


class StatusCommand(Command):
    """
        Retrieve the work order status
    """

    def setup(self, subparsers):
        parser = super(StatusCommand, self).setup(subparsers)
        parser.add_argument('-u', "--api_url", required=True, type=str, help="url of your Customer api")
        parser.add_argument('-i', '--work_order_id', required=True, type=str, help="Work order id")
        return parser

    def run(self, api_url, work_order_id, **kwargs):
        return SiteManager().status_work_order(api_url, work_order_id)



# -*- coding: utf-8 -*-

import os
import sys
import logging
import threading
from tqdm import tqdm
from deepomatic.api.http_helper import HTTPHelper
from deepomatic.cli.cmds.studio_helpers.file import DatasetFiles, UploadImageGreenlet
from deepomatic.cli.cmds.studio_helpers.task import Task
from deepomatic.cli.common import TqdmToLogger, Queue, SUPPORTED_FILE_INPUT_FORMAT
from deepomatic.cli.thread_base import Pool, MainLoop, CurrentMessages
from deepomatic.cli.version import __title__, __version__

###############################################################################

GREENLET_NUMBER = int(os.getenv('DEEPOMATIC_CLI_ADD_IMAGES_CONCURRENCY', 5))
LOGGER = logging.getLogger(__name__)
API_HOST = os.getenv('STUDIO_URL', 'https://studio.deepomatic.com/api/')
DEFAULT_USER_AGENT_PREFIX = user_agent_prefix = '{}/{}'.format(
    __title__, __version__)

###############################################################################


class Client(object):
    def __init__(self, api_key=None, user_agent_prefix=DEFAULT_USER_AGENT_PREFIX, pool_maxsize=GREENLET_NUMBER):
        self.http_helper = HTTPHelper(api_key=api_key, host=API_HOST, user_agent_prefix=user_agent_prefix,
                                      pool_maxsize=pool_maxsize, version=None)
        self.task = Task(self.http_helper)


def get_all_files_with_ext(path, supported_ext, recursive=True):
    """Scans path to find all supported extensions."""
    all_files = []
    if os.path.isfile(path):
        ext_file = os.path.splitext(path)[1].lower()
        if ext_file not in supported_ext:
            LOGGER.warning(
                "The path {} is neither a supported file {} nor a directory, it has been ignored.".format(path, supported_ext))
        else:
            all_files.append(path)
    elif os.path.isdir(path):
        for file in os.listdir(path):
            file_path = os.path.join(path, file)
            if recursive:
                all_files.extend(get_all_files_with_ext(file_path, supported_ext))
            elif os.path.isfile(file_path) and os.path.splitext(file_path)[1].lower() in supported_ext:
                all_files.append(file_path)

    return all_files


def get_all_files(paths, find_txt=False, recursive=True):
    """Retrieves all files from paths, either images or txt if specified."""
    # Make sure path is a list
    paths = [paths] if not isinstance(paths, list) else paths

    # Go through all paths and find corresponding files
    file_ext = ['.txt'] if find_txt else SUPPORTED_FILE_INPUT_FORMAT
    files = []
    for path in paths:
        files += get_all_files_with_ext(path, file_ext, recursive)

    return files


class AddImageManager(object):

    def upload(self, args):
        # Initialize deepomatic client
        clt = Client()

        # Retrieve arguments (must be the same as the parser)
        project_name = args.get('project')
        org_slug = args.get('org')

        paths = args.get('input', [])
        txt_file = args.get('txt_file', False)
        recursive = args.get('recursive', False)
        set_metadata_path = args.get('set_metadata_path', False)

        # Scan to find all files
        files = get_all_files(
            paths=paths, find_txt=txt_file, recursive=recursive)

        # TODO: add a maxsize to avoid taking too much memories
        # This implies reading twice to get the total_files
        queue = Queue()

        dataset_files = DatasetFiles(clt.http_helper, queue)
        total_files = dataset_files.post_files(org_slug, project_name, files)

        exit_event = threading.Event()

        current_messages = CurrentMessages()

        # Initialize progress bar
        tqdmout = TqdmToLogger(LOGGER, level=logging.INFO)
        pbar = tqdm(total=total_files, file=tqdmout,
                    desc='Uploading images', smoothing=0)

        pools = [
            Pool(GREENLET_NUMBER, UploadImageGreenlet,
                 thread_args=(exit_event, queue, current_messages,
                              clt.http_helper, clt.task, pbar.update,
                              set_metadata_path))
        ]

        # Start uploading
        loop = MainLoop(pools, [queue], pbar, exit_event, current_messages)
        try:
            loop.run_forever()
        except Exception:
            loop.cleanup()
            raise

        # If the process encountered an error, the exit code is 1.
        # If the process is interrupted using SIGINT (ctrl + C) or SIGTERM, the threads are stopped, and
        # the exit code is 0.
        if exit_event.is_set():
            sys.exit(1)



from ...utils import PlatformCommand


class EngageAppVersionCommand(PlatformCommand):
    """EngageAppVersion related commands."""

    from .create import CreateCommand
    from .create_from import CreateFromCommand
    from .delete import DeleteCommand



from ...utils import PlatformCommand
from ..utils import DrivePlatformManager


class DeleteCommand(PlatformCommand):
    """Delete a DriveAppVersion."""

    def setup(self, subparsers):
        parser = super(DeleteCommand, self).setup(subparsers)
        parser.add_argument('-v', '--drive_app_version_id', required=True, type=str, help="DriveAppVersion id")
        return parser

    def run(self, drive_app_version_id, **kwargs):
        return DrivePlatformManager().delete_app_version(drive_app_version_id)



import logging
import traceback
import heapq
import gevent
import signal
from contextlib import contextmanager
from gevent.threadpool import ThreadPool
from threading import Lock
from .common import clear_queue, Full, Empty
from deepomatic.api.exceptions import BadStatus
from .exceptions import DeepoCLIException


LOGGER = logging.getLogger(__name__)
QUEUE_MAX_SIZE = 50
SLEEP_TIME = 0.0001  # don't touch until we have non performance regression tests


@contextmanager
def try_lock(lock):
    acquired = lock.acquire(False)
    try:
        yield acquired
    finally:
        if acquired:
            lock.release()


@contextmanager
def blocking_lock(lock, sleep_time=SLEEP_TIME):
    while True:
        # avoid lock totally the main thread and blocking greenlets
        # it may take a little more time to lock if it is not lucky
        # though you can decrease sleep_time to have more luck
        gevent.sleep(sleep_time)
        with try_lock(lock) as acquired:
            if acquired:
                yield
                return


class CurrentMessages(object):
    """
    Track all messages currently being processed in the Pipeline.
    Also allow to track number of errors.
    """
    def __init__(self):
        self.heap_lock = Lock()
        self.messages = []
        self.nb_errors = 0
        self.nb_successes = 0
        self.nb_added_messages = 0

    def lock(self):
        return blocking_lock(self.heap_lock)

    def add_message(self, msg):
        with self.lock():
            heapq.heappush(self.messages, msg)
            self.nb_added_messages += 1

    def get_min(self):
        with self.lock():
            if len(self.messages) > 0:
                return heapq.nsmallest(1, self.messages)[0]
        return None

    def pop_min(self):
        with self.lock():
            if len(self.messages) > 0:
                return heapq.heappop(self.messages)
        return None

    def report_success(self):
        self.report_successes(1)

    def report_successes(self, nb_successes):
        with self.lock():
            self.nb_successes += nb_successes

    def report_error(self):
        self.report_errors(1)

    def report_errors(self, nb_errors):
        with self.lock():
            self.nb_errors += nb_errors

    def report_message(self):
        self.report_messages(1)

    def report_messages(self, nb_messages):
        with self.lock():
            self.nb_added_messages += nb_messages

    def forget_message(self, msg, count_as_error=True):
        try:
            with self.lock():
                if count_as_error:
                    self.nb_errors += 1
                self.messages.remove(msg)
                heapq.heapify(self.messages)
        except ValueError as e:
            # TODO: remove the try/catch
            # we should call it only if we are sure the message it there
            LOGGER.error(str(e))


class ThreadBase(object):
    """
    Thread interface
    """
    def __init__(self, exit_event, input_queue=None, output_queue=None, current_messages=None, name=None):
        super(ThreadBase, self).__init__()
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.exit_event = exit_event
        self.stop_asked = False
        self.name = name or self.__class__.__name__
        # Allow to not stop when an item is still being processed
        self.processing_item_lock = Lock()
        self.current_messages = current_messages
        self.alive = False

    def try_lock(self):
        return try_lock(self.processing_item_lock)

    def can_stop(self):
        if self.input_queue is not None and not self.input_queue.empty():
            return False
        return True

    def stop(self):
        # Can be called from the same thread or from another
        self.stop_asked = True

    def wait_until_nothing_to_process(self):
        # Must be called externally (from another thread)

        if self.input_queue is None:
            # When ThreadBase has no input queue
            # Either the ThreadBase stop by itself by calling self.stop()
            # Either another thread call thread/pool.stop()
            # Otherwise it will wait indefinitely
            self.join()
            return

        # When ThreadBase has an input queue and previous pools are stopped
        # We can stop when the input queue is empty
        long_sleep = 0.05
        sleep_time = long_sleep
        while not self.exit_event.is_set():
            # don't touch until we have non performance regression tests
            gevent.sleep(sleep_time)
            # try to acquire only sometimes
            with self.try_lock() as acquired:
                if acquired:
                    sleep_time = long_sleep
                    if self.can_stop():
                        self.stop()
                        return
                else:
                    sleep_time = SLEEP_TIME
            if not self.alive:
                return

    def process_msg(self, msg):
        raise NotImplementedError()

    def pop_input(self):
        return self.input_queue.get(block=False)

    def put_to_output(self, msg_out):
        while True:
            try:
                self.output_queue.put(msg_out, block=False)
                break
            except Full:
                # don't touch until we have non performance regression tests
                gevent.sleep(SLEEP_TIME)

    def task_done(self, msg_in, msg_out):
        if self.input_queue is not None:
            self.input_queue.task_done()

    def init(self):
        pass

    def close(self):
        pass

    def _run(self):
        while not self.stop_asked:
            empty = False
            with self.try_lock() as acquired:
                if acquired:
                    msg_in = None
                    if self.input_queue is not None:
                        try:
                            msg_in = self.pop_input()
                        except Empty:
                            empty = True

                    if self.input_queue is None or not empty:
                        msg_out = self.process_msg(msg_in)
                        if msg_out is not None:
                            self.put_to_output(msg_out)
                        self.task_done(msg_in, msg_out)
            if empty:
                # don't touch until we have non performance regression tests
                gevent.sleep(SLEEP_TIME)

    def run(self):
        self.alive = True
        try:
            self.init()
            self._run()
        except DeepoCLIException as e:
            LOGGER.error(e)
            self.exit_event.set()
        except BadStatus as e:
            if e.status_code >= 400 and e.status_code < 500:
                LOGGER.error("API raised a bad status code {}: {}".format(
                    e.status_code, e.content
                ))
            else:
                # TODO: we probably want to retry on some errors earlier in the greenlet
                LOGGER.error("Encountered an unexpected exception during routine: {}".format(traceback.format_exc()))
            self.exit_event.set()
        except Exception:
            LOGGER.error("Encountered an unexpected exception during routine: {}".format(traceback.format_exc()))
            self.exit_event.set()
        finally:
            try:
                self.close()
            except Exception:
                LOGGER.error("Encountered an unexpected exception during routine closing: {}".format(traceback.format_exc()))
                self.exit_event.set()
        LOGGER.debug('Quitting {}'.format(self.name))
        self.alive = False


class Thread(ThreadBase):
    def __init__(self, *args, **kwargs):
        super(Thread, self).__init__(*args, **kwargs)
        self.thread = ThreadPool(1)

    def start(self):
        self.thread.spawn(self.run)

    def join(self):
        self.thread.join()


class Greenlet(ThreadBase):
    def __init__(self, *args, **kwargs):
        super(Greenlet, self).__init__(*args, **kwargs)
        self.greenlet = None

    def start(self):
        self.greenlet = gevent.spawn(self.run)

    def join(self):
        if self.greenlet is not None:
            self.greenlet.join()


class Pool(object):
    def __init__(self, nb_thread, thread_cls=Thread, thread_args=(), thread_kwargs=None, name=None):
        self.nb_thread = nb_thread
        self.name = name or thread_cls.__name__
        self.threads = []
        thread_kwargs = thread_kwargs or {}
        for i in range(self.nb_thread):
            th = thread_cls(*thread_args, **thread_kwargs)
            th.name = '{}_{}'.format(self.name, i)
            self.threads.append(th)

    def start(self):
        for th in self.threads:
            th.start()

    def wait_until_nothing_to_process(self):
        # Must be called externally (from another thread)
        for th in self.threads:
            th.wait_until_nothing_to_process()

    def stop(self):
        for th in self.threads:
            th.stop()

    def join(self):
        for th in self.threads:
            th.join()


class MainLoop(object):
    def __init__(self, pools, queues, pbar, exit_event,
                 current_messages, cleanup_func=None):
        self.pools = pools
        self.queues = queues
        self.pbar = pbar
        self.exit_event = exit_event
        self.current_messages = current_messages
        self.cleanup_func = cleanup_func
        self.stop_asked = 0
        self.cleaned = False

    def clear_queues(self):
        # Makes sure all queues are empty
        LOGGER.debug("Purging queues")
        while True:
            for queue in self.queues:
                clear_queue(queue)
            if all([queue.empty() for queue in self.queues]):
                break
        LOGGER.debug("Purging queues done")

    @contextmanager
    def disable_exit_signals(self):
        gevent.signal_handler(signal.SIGINT, lambda: signal.SIG_IGN)
        gevent.signal_handler(signal.SIGTERM, lambda: signal.SIG_IGN)
        try:
            yield
        finally:
            gevent.signal_handler(signal.SIGINT, self.stop)
            gevent.signal_handler(signal.SIGTERM, self.stop)

    def stop(self):
        with self.disable_exit_signals():
            self.stop_asked += 1
            if self.stop_asked < 2:
                LOGGER.info('Stop asked, waiting for threads to process queued messages.')
                # stopping inputs
                self.pools[0].stop()
            elif self.stop_asked == 2:
                LOGGER.info("Hard stop")
                for pool in self.pools:
                    pool.stop()

                # clearing queues to make sure a thread
                # is not blocked in a queue.put() because of maxsize
                self.clear_queues()

    def cleanup(self):
        if self.cleaned:
            return
        if self.cleanup_func is not None:
            self.cleanup_func()

        # Compute the stats on number of errors
        # pbar total may be None for infinite streams
        total_inputs = float('inf') if self.pbar.total is None else self.pbar.total

        nb_uncompleted = (self.current_messages.nb_added_messages
                          - self.current_messages.nb_errors
                          - self.current_messages.nb_successes)
        self.pbar.close()
        LOGGER.info('Summary: errors={} uncompleted={} successful={} total={}.'.format(self.current_messages.nb_errors,
                                                                                       nb_uncompleted,
                                                                                       self.current_messages.nb_successes,
                                                                                       total_inputs))
        self.cleaned = True

    def run_forever(self):
        # Start threads
        for pool in self.pools:
            pool.start()

        gevent.signal_handler(gevent.signal.SIGINT, self.stop)
        gevent.signal_handler(gevent.signal.SIGTERM, self.stop)

        for pool in self.pools:
            # Either pools stop by themself
            # Or they will get stopped when input queue is empty
            pool.wait_until_nothing_to_process()

        if not self.exit_event.is_set():
            gevent.signal_handler(gevent.signal.SIGINT, lambda: signal.SIG_IGN)
            gevent.signal_handler(gevent.signal.SIGTERM, lambda: signal.SIG_IGN)
            # Makes sure threads finish properly so that
            # we can make sure the workflow is not used and can be closed
            for pool in self.pools:
                pool.join()

        self.cleanup()
        return self.stop_asked



import json
import logging
import argparse
import os
import re

from deepomatic.cli.cmds import parser_helpers
from deepomatic.cli.common import (SUPPORTED_IMAGE_INPUT_FORMAT, SUPPORTED_IMAGE_OUTPUT_FORMAT,
                                   SUPPORTED_PROTOCOLS_INPUT, SUPPORTED_VIDEO_INPUT_FORMAT,
                                   SUPPORTED_VIDEO_OUTPUT_FORMAT, SUPPORTED_FOURCC,
                                   SUPPORTED_VIDEO_OUTPUT_COLOR_SPACE)


logger = logging.getLogger(__name__)


class CommandResult:
    """Wrapper arround Command results.

    Attributes:
        resource_name (str)
        extra (str)
        operation (str)
        fields_filter (list)
        data (dict)
    """
    def __init__(self, operation, resource_name, data, fields_filter=None, extra=None):
        self.operation = operation
        self.resource_name = resource_name
        self.data = data
        self.fields_filter = fields_filter or ['id']
        self.extra = extra

    def __repr__(self):
        return str(self)

    def __str__(self):
        important_data = [
            '{}={}'.format(field, self.data[field])
            for field in self.fields_filter
        ]
        message = '[{}] {}: {}'.format(
            self.operation,
            self.resource_name,
            ' '.join(important_data),
        )
        if self.extra:
            message += f' {self.extra}'

        return message

    def to_json_str(self, *args, **kwargs):
        return json.dumps(self.data, *args, **kwargs)


class Command(object):
    """
        Base command, use docstring to fill the command help
        The next lines are used for filling the command description
    """

    def __init__(self):
        # The name of a command is the Command class name lowercase and - joined in case of camel case
        camel_case_name = self.__class__.__name__.replace("Command", "")
        self.name = re.sub(r'(?<!^)(?=[A-Z])', '-', camel_case_name).lower()
        self.help, _, self.description = self.__class__.__doc__.strip().partition('\n')

    def setup(self, subparsers):
        parser = subparsers.add_parser(self.name, help=self.help, description=self.description)
        parser.set_defaults(func=lambda args: self.run(**args))

        subcommands = [command for command in [getattr(self.__class__, attr) for attr in dir(
            self.__class__)] if isinstance(command, type) and issubclass(command, Command)]
        if subcommands:
            subparser = parser.add_subparsers(dest='{}_command'.format(self.name), help='')
            subparser.required = True
            for subcommand in subcommands:
                subcommand().setup(subparser)
        # add verbose
        parser.add_argument('--verbose', dest='verbose', action='store_true',
                            help='Increase output verbosity.')

        return parser

    def run(self, *args, **kwargs):
        print(type(self).__name__, args, kwargs)


class PlatformCommand(Command):
    """Wrapper around Platform Command.

    Add possiblity to format output as json.
    """

    def setup(self, subparsers):
        parser = super().setup(subparsers)
        parser.add_argument(
            '--json-output',
            dest='json_output',
            action='store_true',
            help='Output raw json from api.'
        )
        parser.add_argument(
            '--json-output-indent',
            dest='json_output_indent',
            default=0,
            type=int,
            help='Indentation of json output.'
        )
        return parser


def valid_path(file_path):
    if not os.path.exists(file_path):
        raise argparse.ArgumentTypeError("'{}' file does not exist".format(file_path))
    return file_path


def valid_json(data):
    try:
        return json.loads(data)
    except Exception:
        raise argparse.ArgumentTypeError("'{}' is not a valid JSON".format(data))


class BuildDict(argparse.Action):
    """
    This class is used in argparse. It will transform a chain of name:values into a dict.
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(BuildDict, self).__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        dic = {}
        for kv in values:
            vals = kv.split(":")
            assert len(vals) == 2
            name = vals[0]
            value = vals[1]
            dic[name] = value
        setattr(namespace, self.dest, dic)


def setup_model_cmd_line_parser(mode, cmd, inference_parsers):
    assert mode in ['site', 'platform']
    assert cmd in ['infer', 'draw', 'blur', 'noop']

    # Define input group for infer draw blur noop
    if cmd in ['infer', 'draw', 'blur', 'noop']:
        # Define argument groups for easier reading
        group = parser_helpers.add_common_cmd_group(inference_parsers, 'input')
        group.add_argument('-i', '--input', required=True,
                           help="Input path, either an image (*{}), a video (*{}), a directory, a stream (*{}),"
                           " or a Studio format (*.txt). If the given path is a directory,"
                           " it will recursively run inference on all the supported files"
                           " in this directory if the -R option is used.".format(', *'.join(SUPPORTED_IMAGE_INPUT_FORMAT),
                                                                                 ', *'.join(SUPPORTED_VIDEO_INPUT_FORMAT),
                                                                                 ', *'.join(SUPPORTED_PROTOCOLS_INPUT)))
        group.add_argument('--input_fps', type=int, help="FPS used for input video frame skipping and extraction."
                           " If higher than the original video FPS, all frames will be analysed only once having"
                           " the same effect as not using this parameter. If lower than the original video FPS,"
                           " some frames will be discarded to simulate an input of the given FPS.", default=None)
        group.add_argument('--skip_frame', type=int, help="Number of frame to skip between two frames from the input."
                           " It can be combined with input_fps", default=0)
        parser_helpers.add_recursive_argument(group)

    output_groups = {}
    # Define output group for infer draw blur noop
    if cmd in ['infer', 'draw', 'blur', 'noop']:
        group = parser_helpers.add_common_cmd_group(inference_parsers, 'output')
        output_groups[cmd] = group
        group = output_groups[cmd]
        group.add_argument('-o', '--outputs', required=True, nargs='+', help="Output path, either an image (*{}),"
                           " a video (*{}), a json (*.json) or a directory.".format(', *'.join(SUPPORTED_IMAGE_OUTPUT_FORMAT),
                                                                                    ', *'.join(SUPPORTED_VIDEO_OUTPUT_FORMAT)))
        group.add_argument('--output_fps', type=int, help="FPS used for output video reconstruction.", default=None)
        group.add_argument('--fourcc', type=str, help="Codec used for output video reconstruction.",
                           choices=set([fourcc for fourccs in SUPPORTED_FOURCC.values() for fourcc in fourccs]), default=None)
        group.add_argument('--output_color_space', type=str,
                           help="Mainly useful for option `-o stdout`. Convert the outputed frame into the specified color space.",
                           choices=SUPPORTED_VIDEO_OUTPUT_COLOR_SPACE, default='BGR')

    # Define output group for draw blur noop
    if cmd in ['draw', 'blur', 'noop']:
        group = output_groups[cmd]
        group.add_argument('-F', '--fullscreen', help="Fullscreen if window output.", action="store_true")

    # Define option group for draw blur
    if cmd in ['draw', 'blur']:
        subparser = inference_parsers
        subparser.add_argument('--from_file', type=str, dest='pred_from_file',
                               help="Uses prediction from a Vulcan JSON")

    # Define model group for infer draw blur
    if cmd in ['infer', 'draw', 'blur']:
        group = inference_parsers.add_argument_group('model arguments')
        reco_required = (mode == 'platform')
        group.add_argument('-r', '--recognition_id', required=reco_required,
                           help="Neural network recognition version ID.")
        group.add_argument('-t', '--threshold', type=float,
                           help="Threshold above which a prediction is considered valid.",
                           default=None)

    # Define onprem group for infer draw blur
    if mode == "site" and cmd in ['infer', 'draw', 'blur']:
        group = inference_parsers.add_argument_group('on-premises arguments')
        group.add_argument('-u', '--amqp_url', required=True,
                           help="AMQP url for on-premises deployments.")
        group.add_argument('-k', '--routing_key', required=True,
                           help="Recognition routing key for on-premises deployments.")

    if cmd == "draw":
        # Define draw specific options
        group = inference_parsers.add_argument_group('drawing arguments')
        group.add_argument('-fs', '--font_scale', dest='font_scale', type=float,
                           help="Text font scale, defaults to 0.5",
                           default=0.5)
        group.add_argument('-ft', '--font_thickness', dest='font_thickness', type=int,
                           help="Text font thickness, must be an int and defaults to 1",
                           default=1)
        group.add_argument('--font_bg_color', default=None, nargs=3,
                           help="Expect a B G R value. If set, draws labels with a unique background color."
                                " By default, the background is red/orange/green depending on the threshold set"
                                " and the prediction score.", type=int)
        score_group = group.add_mutually_exclusive_group()
        score_group.add_argument('-S', '--draw_scores', dest='draw_scores', help="Overlay the prediction scores. Default behavior.",
                                 action="store_true")
        score_group.add_argument('--no_draw_scores', dest='draw_scores', help="Do not overlay the prediction scores.", action="store_false")
        score_group.set_defaults(draw_scores=True)
        label_group = group.add_mutually_exclusive_group()
        label_group.add_argument('-L', '--draw_labels', dest='draw_labels', help="Overlay the prediction labels. Default behavior.",
                                 action="store_true")
        label_group.add_argument('--no_draw_labels', dest='draw_labels', help="Do not overlay the prediction labels.", action="store_false")
        label_group.set_defaults(draw_labels=True)

    if cmd == "blur":
        # Define blur specific options
        group = inference_parsers.add_argument_group('blurring arguments')
        group.add_argument('-M', '--blur_method', help="Blur method to apply, either 'pixel', 'gaussian' or 'black', defaults to 'pixel'.",
                           default='pixel', choices=['pixel', 'gaussian', 'black'])
        group.add_argument('-B', '--blur_strength', help="Blur strength, defaults to 10.", default=10)



from .utils import SiteManager, _SiteCommand


class UseCommand(_SiteCommand):
    """
        Use a site
    """

    def run(self, site_id, **kwargs):
        if SiteManager().use(site_id):
            print('Now using', site_id)
        else:
            print(site_id, 'is not installed')



__title__ = 'deepomatic-cli'
__description__ = 'Deepomatic CLI'
__version__ = '0.7.5'
__author__ = 'deepomatic'
__author_email__ = 'support@deepomatic.com'
__url__ = 'https://github.com/Deepomatic/deepocli'
__license__ = 'MIT License'



from ...utils import Command, BuildDict
from ..utils import SiteManager


class CreateCommand(Command):
    """
        Create a new work order
    """

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-u', "--api_url", required=True, type=str, help="url of your Customer api")
        parser.add_argument('-n', '--name', required=True, type=str, help="Work order name")
        parser.add_argument('-m', "--metadata", dest='metadata',
                            action=BuildDict, nargs="+", metavar="NAME:VAL", required=True,
                            help='Metadata for the work order in format name:value')
        return parser

    def run(self, api_url, name, metadata, **kwargs):
        return SiteManager().create_work_order(api_url, name, metadata)



from ...utils import PlatformCommand
from ..utils import EngagePlatformManager


class CreateCommand(PlatformCommand):
    """Create a new EngageApp."""

    def setup(self, subparsers):
        parser = super(CreateCommand, self).setup(subparsers)
        parser.add_argument('-n', '--name', required=True, type=str, help="EngageApp name")
        parser.add_argument(
            '-t',
            '--application_type',
            type=str,
            help="Engage App type (INFERENCE, WORKFLOW, FIELD_SERVICES, VIDEO). Default to FIELD_SERVICES."
        )
        return parser

    def run(self, name, application_type, **kwargs):
        return EngagePlatformManager().create_app(name, application_type)
